# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for language model."""

import abc
import contextlib
import dataclasses
import datetime
import enum
import functools
import math
import re
import threading
import time
from typing import Annotated, Any, Callable, ClassVar, Iterator, Literal, Optional, Sequence, Tuple, Type, Union, final
from langfun.core import async_support
from langfun.core import component
from langfun.core import concurrent
from langfun.core import console
from langfun.core import message as message_lib

import pyglove as pg

#
# Common errors during calling language models.
#


class LMError(RuntimeError):
  """Base class for language model errors."""


class LMInputError(LMError):
  """Base class for errors with bad input."""


class ContextLimitError(LMInputError):
  """Error for context limit exceeded."""


class RetryableLMError(LMError):
  """Base class for LLM errors that can be solved by retrying."""


class RateLimitError(RetryableLMError):
  """Error for rate limit reached."""


class TemporaryLMError(RetryableLMError):
  """Error for temporary service issues that can be retried."""


#
# Language model information.
#


class ModelInfo(pg.Object):
  """Common information for a language model."""

  # Constant for modalities.
  TEXT_INPUT_ONLY = []

  model_id: Annotated[
      str,
      'A global unique identifier of the language model. ',
  ]

  alias_for: Annotated[
      str | None,
      'The fixed-version model ID that this model is aliased for.',
  ] = None

  #
  # Basic information.
  #

  model_type: Annotated[
      Literal['unknown', 'pretrained', 'instruction-tuned', 'thinking'],
      'The type of the model.'
  ] = 'unknown'

  provider: Annotated[
      str | None,
      (
          'The service provider (host) of the LLM. E.g. VertexAI, Microsoft, '
          'etc.'
      )
  ] = None

  description: Annotated[
      str | None,
      'An optional description of the model family.'
  ] = None

  url: Annotated[
      str | None,
      'The URL of the model.'
  ] = None

  release_date: Annotated[
      datetime.date | None,
      'The release date of the model. '
  ] = None

  in_service: Annotated[
      bool,
      'If True, the model is in service.'
  ] = True

  #
  # LLM capabilities.
  #

  input_modalities: Annotated[
      list[str] | None,
      (
          'Supported MIME types as model inputs. '
          'If None, this information is unknown, so Langfun allows all '
          'modalities from the input to be passed to the model.'
      )
  ] = None

  class ContextLength(pg.Object):
    """Context length information."""

    max_input_tokens: Annotated[
        int | None,
        (
            'The maximum number of input tokens of the language model. '
            'If None, there is no limit or this information is unknown.'
        )
    ] = None

    max_output_tokens: Annotated[
        int | None,
        (
            'The maximum number of output tokens of the language model. '
            'If None, there is no limit or this information is unknown.'
        )
    ] = None

    max_cot_tokens: Annotated[
        int | None,
        (
            'The maximum number of Chain-of-Thought tokens to generate. '
            'If None, there is not limit or not applicable.'
        )
    ] = None

  context_length: Annotated[
      ContextLength | None,
      (
          'Context length information of the model. '
          'If None, this information is unknown.'
      )
  ] = None

  #
  # Common pricing information.
  #

  class Pricing(pg.Object):
    """Pricing information."""

    cost_per_1m_cached_input_tokens: Annotated[
        float | None,
        (
            'The cost per 1M cached input tokens in US dollars. '
            'If None, this information is unknown.'
        )
    ] = None

    cost_per_1m_input_tokens: Annotated[
        float | None,
        (
            'The cost per 1M input tokens in US dollars. '
            'If None, this information is unknown.'
        )
    ] = None

    cost_per_1m_output_tokens: Annotated[
        float | None,
        (
            'The cost per 1M output tokens in US dollars. '
            'If None, this information is unknown.'
        )
    ] = None

    def estimate_cost(self, usage: 'LMSamplingUsage') -> float | None:
      """Estimates the cost of using the model. Subclass could override.

      Args:
        usage: The usage information of the model.

      Returns:
        The estimated cost in US dollars. If None, cost estimating is not
        supported on the model.
      """
      # NOTE(daiyip): supported cached tokens accounting in future.
      if (self.cost_per_1m_input_tokens is None
          or self.cost_per_1m_output_tokens is None):
        return None
      return (
          self.cost_per_1m_input_tokens * usage.prompt_tokens
          + self.cost_per_1m_output_tokens
          * (usage.total_tokens - usage.prompt_tokens)
      ) / 1000_000

  pricing: Annotated[
      Pricing | None,
      (
          'Pricing information. If None, this information is unknown.'
      )
  ] = None

  #
  # Rate limits.
  #

  class RateLimits(pg.Object):
    """Preset rate limits."""

    max_requests_per_minute: Annotated[
        int | None,
        (
            'The max number of requests per minute.'
            'If None, there is no limit.'
        )
    ] = None

    max_tokens_per_minute: Annotated[
        int | None,
        (
            'The max number of tokens per minute.'
            'If None, there is no limit.'
        )
    ] = None

  rate_limits: Annotated[
      RateLimits | None,
      (
          'Rate limits. If None, this information is unknown.'
      )
  ] = None

  #
  # Additional information.
  #

  metadata: Annotated[
      dict[str, Any],
      (
          'Model metadata. This could be used to store model-specific '
          'information, which could be consumed by modules that need to '
          'apply model-specific logic.'
      ),
  ] = {}

  #
  # Common model protocols.
  #

  @final
  def estimate_cost(self, usage: 'LMSamplingUsage') -> float | None:
    """Estimates the cost of using the model."""
    if self.pricing is None:
      return None
    return self.pricing.estimate_cost(usage)

  def supports_input(self, mime_type: str) -> bool:
    """Returns True if an input MIME type is supported.

    Subclass could override.

    Args:
      mime_type: The MIME type of the input.

    Returns:
      True if the input MIME type is supported.
    """
    if self._input_modalities is None:
      return True
    return mime_type.lower() in self._input_modalities

  @property
  def resource_id(self) -> str:
    """Returns the resource ID of the LLM. Subclass could override."""
    canonical_model_id = self.alias_for or self.model_id
    if self.provider is None or not pg.is_deterministic(self.provider):
      return canonical_model_id
    provider = self.provider.lower().replace(' ', '_')
    return f'{provider}://{canonical_model_id}'

  def _on_bound(self):
    super()._on_bound()
    self._input_modalities = set(
        [mime_type.lower() for mime_type in self.input_modalities]
    ) if self.input_modalities is not None else None


#
# Language model input/output interfaces.
#


class LMSample(pg.Object):
  """Response candidate."""

  response: pg.typing.Annotated[
      pg.typing.Object(
          message_lib.Message,
          # Allowing automatic conversion from text to AIMessage.
          transform=message_lib.AIMessage.from_value,
      ),
      'The natural language response of LM.',
  ]

  score: Annotated[
      float, 'The score of sampled response. The larger is better'
  ] = 0.0

  logprobs: Annotated[
      list[tuple[str, float, list[tuple[str, float]]]] | None,
      '(token, log prob, top tokens and their probs).',
  ] = None


class RetryStats(pg.Object):
  """Retry stats, which is aggregated across multiple retry entries."""

  num_occurences: Annotated[
      int,
      'Total number of retry attempts on LLM (excluding the first attempt).',
  ] = 0

  total_wait_interval: Annotated[
      float, 'Total wait interval in seconds due to retry.'
  ] = 0

  total_call_interval: Annotated[
      float, 'Total LLM call interval in seconds.'
  ] = 0

  errors: Annotated[
      dict[str, int],
      'A Counter of error types encountered during the retry attempts.',
  ] = {}

  @classmethod
  def from_retry_entries(
      cls, retry_entries: Sequence[concurrent.RetryEntry]
  ) -> 'RetryStats':
    """Creates a RetryStats from a sequence of RetryEntry."""
    if not retry_entries:
      return RetryStats()
    errors = {}
    for retry in retry_entries:
      if retry.error is not None:
        errors[retry.error.__class__.__name__] = (
            errors.get(retry.error.__class__.__name__, 0) + 1
        )
    return RetryStats(
        num_occurences=len(retry_entries) - 1,
        total_wait_interval=sum(e.wait_interval for e in retry_entries),
        total_call_interval=sum(e.call_interval for e in retry_entries),
        errors=errors,
    )

  def __add__(self, other: 'RetryStats') -> 'RetryStats':
    errors = self.errors.copy()
    for error, count in other.errors.items():
      errors[error] = errors.get(error, 0) + count
    return RetryStats(
        num_occurences=self.num_occurences + other.num_occurences,
        total_wait_interval=self.total_wait_interval
        + other.total_wait_interval,
        total_call_interval=self.total_call_interval
        + other.total_call_interval,
        errors=errors,
    )

  def __radd__(self, other: 'RetryStats') -> 'RetryStats':
    return self + other


class LMSamplingUsage(pg.Object):
  """Usage information per completion."""

  prompt_tokens: int
  completion_tokens: int
  total_tokens: int
  num_requests: int = 1
  estimated_cost: Annotated[
      float | None,
      (
          'Estimated cost in US dollars. If None, cost estimating is not '
          'suppported on the model being queried.'
      ),
  ] = None
  retry_stats: RetryStats = RetryStats()
  completion_tokens_details: dict[str, Any] | None = None

  def __bool__(self) -> bool:
    return self.num_requests > 0

  @property
  def average_prompt_tokens(self) -> int:
    """Returns the average prompt tokens per request."""
    return self.prompt_tokens // self.num_requests

  @property
  def average_completion_tokens(self) -> int:
    """Returns the average completion tokens per request."""
    return self.completion_tokens // self.num_requests

  @property
  def average_total_tokens(self) -> int:
    """Returns the average total tokens per request."""
    return self.total_tokens // self.num_requests

  @property
  def average_estimated_cost(self) -> float | None:
    """Returns the average estimated cost per request."""
    if self.estimated_cost is None:
      return None
    return self.estimated_cost / self.num_requests

  def __add__(self, other: Optional['LMSamplingUsage']) -> 'LMSamplingUsage':
    if other is None:
      return self
    if self.estimated_cost is None:
      estimated_cost = other.estimated_cost
    elif other.estimated_cost is None:
      estimated_cost = self.estimated_cost
    else:
      estimated_cost = self.estimated_cost + other.estimated_cost
    return LMSamplingUsage(
        prompt_tokens=self.prompt_tokens + other.prompt_tokens,
        completion_tokens=self.completion_tokens + other.completion_tokens,
        total_tokens=self.total_tokens + other.total_tokens,
        num_requests=self.num_requests + other.num_requests,
        estimated_cost=estimated_cost,
        retry_stats=self.retry_stats + other.retry_stats,
    )

  def __radd__(self, other: Optional['LMSamplingUsage']) -> 'LMSamplingUsage':
    return self + other


class UsageNotAvailable(LMSamplingUsage):
  """Usage information not available."""
  prompt_tokens: pg.typing.Int(0).freeze()       # pytype: disable=invalid-annotation
  completion_tokens: pg.typing.Int(0).freeze()   # pytype: disable=invalid-annotation
  total_tokens: pg.typing.Int(0).freeze()        # pytype: disable=invalid-annotation
  estimated_cost: pg.typing.Float(default=None, is_noneable=True).freeze()    # pytype: disable=invalid-annotation

  def __add__(self, other: Optional['LMSamplingUsage']) -> 'UsageNotAvailable':
    if other is None:
      return self
    return UsageNotAvailable(
        num_requests=self.num_requests + other.num_requests
    )

  def __radd__(self, other: Optional['LMSamplingUsage']) -> 'UsageNotAvailable':
    return self + other


class LMSamplingResult(pg.Object):
  """Language model response."""

  samples: Annotated[
      list[LMSample],
      (
          'Multiple samples of the same prompt, sorted by their score. '
          'The first candidate has the highest score.'
      ),
  ] = []

  usage: Annotated[
      LMSamplingUsage,
      'Usage information. Currently only OpenAI models are supported.',
  ] = UsageNotAvailable()

  is_cached: Annotated[
      bool,
      'Whether the result is from cache or not.'
  ] = False


class LMSamplingOptions(component.Component):
  """Language model sampling options."""

  temperature: Annotated[
      float | None,
      (
          'Model temperature, which is usually between 0 and 1.0. '
          'OpenAI models have temperature range from 0.0 to 2.0. '
          'If None (default), honor the model\'s default behavior. '
      )
  ] = None

  max_tokens: Annotated[
      int | None,
      (
          'Per example max tokens to generate. '
          'If None, use the model default.'
      )
  ] = None

  n: Annotated[int | None, 'Max number of samples to return.'] = 1

  top_k: Annotated[
      int | None,
      (
          'Top k tokens to sample the next token. '
          'Not applicable to OpenAI models.'
      )
  ] = 40

  top_p: Annotated[
      float | None,
      (
          'Only sample the next token from top N tokens whose accumulated '
          'probability // mass <= p. For OpenAI models, set `temperature` or '
          '`top_p` but not both.'
      ),
  ] = None

  stop: Annotated[
      list[str] | None,
      (
          'A list of stop sequences that prevent LLMs from outputting '
          'more tokens. For example, when `stop` is set to ["User:", "Model:"] '
          'LLMs will stop to emit more tokens when `User:` or '
          '`Model:` is reached.'
      ),
  ] = None

  random_seed: Annotated[
      int | None, 'A fixed random seed used during model inference.'
  ] = None

  logprobs: Annotated[
      bool,
      (
          'Whether to return log probabilities of the output tokens or not. If '
          'true, returns the log probabilities of each output token returned '
          'in the content of message.'
      ),
  ] = False

  top_logprobs: Annotated[
      int | None,
      (
          'An integer between 0 and 5 specifying the number of most likely '
          'tokens to return at each token position, each with an associated '
          'log probability. logprobs must be set to true if this parameter is '
          'used.'
      ),
  ] = None

  max_thinking_tokens: Annotated[
      int | None, 'Number of max thinking tokens.'
  ] = None

  reasoning_effort: Annotated[
      Literal['low', 'medium', 'high'] | None,
      (
          'This parameter is used by OpenAI reasoning models (e.g. O3, O4 mini)'
          ' to guides how many reasoning tokens to generate before creating a'
          ' response to the prompt.'
      ),
  ] = None

  def cache_key(self) -> tuple[Any, ...]:
    """Returns a tuple of current values as cache key."""
    return (
        self.temperature,
        self.max_tokens,
        self.n,
        self.top_k,
        self.top_p,
        self.random_seed
    )


class LMScoringResult(pg.Object):
  """Language model scoring result."""

  score: Annotated[
      float,
      'The log likelyhood of the requested completion towards the prompt.',
  ]
  gradients: Annotated[
      Any | None,
      '(Optional) gradients from the score method, w.r.t.' +
      ' prompt.metadata.weights.',
  ] = None


class LMCache(pg.Object):
  """Interface for LM cache."""

  @dataclasses.dataclass
  class Stats:
    """Cache stats."""

    num_queries: int = 0
    num_hits: int = 0
    num_hit_expires: int = 0
    num_misses: int = 0
    num_updates: int = 0
    num_deletes: int = 0

  @abc.abstractmethod
  def get(
      self, lm: 'LanguageModel', prompt: message_lib.Message, seed: int
  ) -> LMSamplingResult | None:
    """Gets the cached result of a prompt generated by a language model."""

  @abc.abstractmethod
  def put(
      self,
      lm: 'LanguageModel',
      prompt: message_lib.Message,
      result: LMSamplingResult,
      seed: int,
  ) -> None:
    """Puts the result of a prompt generated by a language model in cache."""

  @abc.abstractmethod
  def delete(
      self,
      lm: 'LanguageModel',
      prompt: message_lib.Message,
      seed: int,
  ) -> bool:
    """Deletes the result of a prompt generated by a language model in cache."""

  @property
  @abc.abstractmethod
  def stats(self) -> Stats:
    """Returns cache stats."""


class LMDebugMode(enum.IntFlag):
  """Sets debugging mode for a language model.

  INFO toggles whether information about the LM will be printed.
  PROMPT toggles whether the prompts sent to the LM will be printed.
  RESPONSE toggles whether the responses from the LM will be printed.
  """
  NONE = 0

  INFO = enum.auto()
  PROMPT = enum.auto()
  RESPONSE = enum.auto()

  ALL = INFO | PROMPT | RESPONSE


class LanguageModel(component.Component):
  """Interface of a language model.

  Language models are at the center of LLM-based agents. ``LanguageModel``
  is the interface to interact with different language modles.

  In langfun, users can use different language models with the same agents,
  allowing fast prototype, as well as side-by-side comparisons.
  """

  sampling_options: LMSamplingOptions = LMSamplingOptions()

  cache: Annotated[
      LMCache | None,
      (
          'Sampling cache. If None, no cache will be used.'
      )
  ] = component.contextual(default=None)

  max_concurrency: Annotated[
      int | None,
      (
          'Max concurrent requests being sent to the server. '
          'If None, there is no limit. '
          'Please note that the concurrency control is based on the '
          '`info.resource_id` property, meaning that model instances shared '
          'the same resource ID will be accounted under the same concurrency '
          'control key. This allows a process-level concurrency control '
          'for specific models regardless the number of LM (client) instances '
          'created by the program. Subclasses could override the '
          '`max_concurrency` property to allow dynamic concurrency control.'
      ),
  ] = None

  timeout: Annotated[
      float | None, 'Timeout in seconds. If None, there is no timeout.'
  ] = 120.0

  max_attempts: Annotated[
      int,
      (
          'A number of max attempts to request the LM if fails.'
          'The retry wait time is determined per LM serivice.'
      ),
  ] = 5

  retry_interval: Annotated[
      int | tuple[int, int],
      (
          'An integer as a constant wait time in seconds before next retry, '
          'or a tuple of two integers representing the range of wait time, '
          'based on which the next wait time will be randmly chosen.'
      )
  ] = (5, 60)

  exponential_backoff: Annotated[
      bool,
      (
          'If True, the wait time among multiple attempts will exponentially '
          'grow. If `retry_interval` is an integer, the wait time for the '
          'k\'th attempt will be `retry_interval * 2 ^ (k - 1)` seconds. If '
          '`retry_interval` is a tuple, the wait time range for the k\'th '
          'attempt will be `(retry_interval[0] * 2 ^ (k - 1), '
          'retry_interval[1] * 2 ^ (k - 1)`) seconds.'
      )
  ] = True

  max_retry_interval: Annotated[
      int,
      (
          'The max retry interval in seconds. This is useful when the retry '
          'interval is exponential, to avoid the wait time to grow '
          'exponentially.'
      )
  ] = 300

  debug: Annotated[
      bool | LMDebugMode,
      (
          'If True, the prompt and the response will be output to stdout. '
          'Specific debugging fields (info, prompt, response) can be specified '
          'using the LMDebugMode flags.'
      ),
  ] = False

  _MODEL_FACTORY: ClassVar[dict[str, Callable[..., 'LanguageModel']]] = {}

  @classmethod
  def register(
      cls,
      model_id_or_prefix: str, factory: Callable[..., 'LanguageModel']
  ) -> None:
    """Registers a factory function for a model ID."""
    cls._MODEL_FACTORY[model_id_or_prefix] = factory

  @classmethod
  def get(cls, model_str: str, *args, **kwargs) -> 'LanguageModel':
    """Creates a language model instance from a model str.

    Args:
      model_str: A string that identifies the model. It can be a model ID or a
        model ID with kwargs.
        For example, "gpt-o3?temperature=0.1&n=2" will create a GPT-o3 model
        with temperature set to 0.1 and n set to 2.
      *args: Additional arguments to pass to the model factory.
      **kwargs: Additional keyword arguments to pass to the model factory.
        kwargs provided here will take precedence over kwargs parsed from
        model_str.

    Returns:
      A language model instance.
    """
    model_id, model_kwargs = cls._parse_model_str(model_str)
    model_kwargs.update(kwargs)

    factory = cls._MODEL_FACTORY.get(model_id)
    if factory is None:
      factories = []
      for k, v in cls._MODEL_FACTORY.items():
        if re.match(k, model_id):
          factories.append((k, v))
      if not factories:
        raise ValueError(f'Model not found: {model_id!r}.')
      elif len(factories) > 1:
        raise ValueError(
            f'Multiple models found for {model_id!r}: '
            f'{[x[0] for x in factories]}. '
            'Please specify a more specific model ID.'
        )
      factory = factories[0][1]
    return factory(model_id, *args, **model_kwargs)

  @classmethod
  def _parse_model_str(cls, model_str: str) -> tuple[str, dict[str, Any]]:
    """Parses a model string into model ID and kwargs."""
    parts = model_str.split('?')
    if len(parts) == 1:
      return model_str, {}
    elif len(parts) == 2:
      model_id, kwargs_str = parts
      kwargs = {}
      for kv in kwargs_str.split('&'):
        kv_parts = kv.split('=')
        if len(kv_parts) != 2:
          raise ValueError(f'Invalid kwargs in model string: {model_str!r}.')
        k, v = kv_parts
        if v.isnumeric():
          v = int(v)
        elif v.lower() in ('true', 'false'):
          v = v.lower() == 'true'
        else:
          v = v.strip()
          try:
            v = float(v)
          except ValueError:
            pass
        kwargs[k] = v
      return model_id, kwargs
    else:
      raise ValueError(f'Invalid model string: {model_str!r}.')

  @classmethod
  def dir(cls, regex: str | None = None):
    """Returns a list of model IDs that match the given regex."""
    if regex is None:
      return sorted(list(LanguageModel._MODEL_FACTORY.keys()))
    return sorted(
        [k for k in LanguageModel._MODEL_FACTORY.keys() if re.match(regex, k)]
    )

  @pg.explicit_method_override
  def __init__(self, *args, **kwargs) -> None:
    """Overrides __init__ to pass through **kwargs to sampling options."""

    sampling_options = kwargs.pop(
        'sampling_options',
        pg.clone(self.__schema__.fields['sampling_options'].default_value)
    )
    sampling_options_delta = {}

    for k, v in kwargs.items():
      if LMSamplingOptions.__schema__.get_field(k) is not None:
        sampling_options_delta[k] = v

    if sampling_options_delta:
      sampling_options.rebind(sampling_options_delta)

    for k in sampling_options_delta:
      del kwargs[k]

    super().__init__(*args, sampling_options=sampling_options, **kwargs)

  def _on_bound(self):
    super()._on_bound()
    self._call_counter = 0
    self.__dict__.pop('model_info', None)

  @functools.cached_property
  def model_info(self) -> ModelInfo:
    """Returns the specification of the model."""
    return ModelInfo(model_id='unknown')

  #
  # Shortcut properties/methods from `model_info`.
  # If these behaviors need to be changed, please override the corresponding
  # methods in the ModelInfo subclasses instead of these properties/methods.
  #

  @final
  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return self.model_info.alias_for or self.model_info.model_id

  @final
  @property
  def resource_id(self) -> str:
    """Resource ID for performing request parallism control."""
    return self.model_info.resource_id

  @final
  @property
  def context_length(self) -> ModelInfo.ContextLength | None:
    """Returns the context length of the model."""
    return self.model_info.context_length

  @final
  @property
  def pricing(self) -> ModelInfo.Pricing | None:
    """Returns the pricing information of the model."""
    return self.model_info.pricing

  @final
  @property
  def rate_limits(self) -> ModelInfo.RateLimits | None:
    """Returns the rate limits to the model."""
    return self.model_info.rate_limits

  @final
  def supports_input(self, mime_type: str):
    """Returns True if an input type is supported. Subclasses can override."""
    return self.model_info.supports_input(mime_type)

  @final
  def estimate_cost(self, usage: LMSamplingUsage) -> float | None:
    """Returns the estimated cost of a usage. Subclasses can override."""
    return self.model_info.estimate_cost(usage)

  #
  # Language model operations.
  #

  async def asample(
      self,
      prompts: list[str | message_lib.Message],
      *,
      cache_seed: int = 0,
      **kwargs,
  ) -> message_lib.Message:
    """Async version of sample."""
    # TODO(daiyip): implement native async sampling.
    return await async_support.invoke_async(
        self.sample, prompts, cache_seed=cache_seed, **kwargs
    )

  async def acall(
      self,
      prompt: str | message_lib.Message,
      *,
      cache_seed: int = 0,
      **kwargs
  ) -> message_lib.Message:
    """Async version of __call__."""
    # TODO(daiyip): implement native async calling.
    return await async_support.invoke_async(
        self.__call__,
        prompt,
        cache_seed=cache_seed,
        **kwargs
    )

  async def ascore(
      self,
      prompt: str | message_lib.Message | list[message_lib.Message],
      completions: list[str | message_lib.Message],
      **kwargs,
  ) -> float:
    """Async version of score."""
    # TODO(daiyip): implement native async scoring.
    return await async_support.invoke_async(
        self.score,
        prompt,
        completions=completions,
        **kwargs
    )

  async def atokenize(
      self,
      prompt: str | message_lib.Message,
      **kwargs
  ) -> list[int]:
    """Async version of tokenize."""
    # TODO(daiyip): implement native async tokenization.
    return await async_support.invoke_async(self.tokenize, prompt, **kwargs)

  def sample(
      self,
      prompts: list[str | message_lib.Message],
      *,
      cache_seed: int = 0,
      **kwargs,
  ) -> list[LMSamplingResult]:
    """Samples one or multiple prompts."""
    # Internal usage logging.

    prompts = [message_lib.UserMessage.from_value(p) for p in prompts]

    with component.context(override_attrs=True, **kwargs):
      if self.cache is None:
        results = self._sample(prompts)
      else:
        results = self._sample_with_cache_lookup(prompts, cache_seed)

      for prompt, result in zip(prompts, results):

        # Tag LM input.
        prompt.tag(message_lib.Message.TAG_LM_INPUT)

        for sample in result.samples:
          # Update metadata for response message.

          response = sample.response
          response.metadata.score = sample.score
          response.metadata.logprobs = sample.logprobs
          response.metadata.is_cached = result.is_cached

          # Update estimated cost.
          usage = result.usage
          estimated_cost = self.estimate_cost(usage)
          if estimated_cost is not None:
            usage.rebind(
                estimated_cost=estimated_cost, skip_notification=True
            )

          # NOTE(daiyip): Current usage is computed at per-result level,
          # which is accurate when n=1. For n > 1, we average the usage across
          # multiple samples.
          if len(result.samples) == 1 or isinstance(usage, UsageNotAvailable):
            response.metadata.usage = usage
          else:
            n = len(result.samples)
            response.metadata.usage = LMSamplingUsage(
                prompt_tokens=usage.prompt_tokens // n,
                completion_tokens=usage.completion_tokens // n,
                total_tokens=usage.total_tokens // n,
                estimated_cost=(
                    usage.estimated_cost / n if usage.estimated_cost else None
                ),
                retry_stats=RetryStats(
                    num_occurences=usage.retry_stats.num_occurences // n,
                    total_wait_interval=usage.retry_stats.total_wait_interval
                    / n,
                    total_call_interval=usage.retry_stats.total_call_interval
                    / n,
                    errors={
                        error: count // n
                        for error, count in usage.retry_stats.errors.items()
                    },
                ),
            )

          # Track usage.
          trackers = component.context_value('__usage_trackers__', [])
          if trackers:
            model_id = self.model_id
            for tracker in trackers:
              tracker.track(model_id, usage, result.is_cached)

          # Track the prompt for corresponding response.
          response.source = prompt

          # Tag LM response.
          response.tag(message_lib.Message.TAG_LM_RESPONSE)
      return results

  def _sample_with_cache_lookup(
      self, prompts: list[str | message_lib.Message], cache_seed: int
  ) -> list[LMSamplingResult]:
    """Sample with cache lookup."""
    assert self.cache is not None

    results = [None] * len(prompts)
    requests, request_to_result_index = [], {}

    # Perform cache lookup and figure out sampling requests to make.
    for i, prompt in enumerate(prompts):
      r = None
      # Query cache if cache_seed is not None.
      if cache_seed is not None:
        r = self.cache.get(self, prompt, seed=cache_seed)

      if r is None:
        request_to_result_index[len(requests)] = i
        requests.append(prompt)
      else:
        result = r.clone()
        assert result.is_cached, result
        results[i] = result

    # Sample non-cache-hit prompts.
    if requests:
      requested_results = self._sample(requests)
      assert len(requested_results) == len(requests), (
          requests, requested_results)

      # Combine cached results and newly requested results.
      for i, (prompt, result) in enumerate(zip(requests, requested_results)):
        results[request_to_result_index[i]] = result

        # Carry the cache seed in response message.
        for sample in result.samples:
          sample.response.set('cache_seed', cache_seed)

        if cache_seed is not None:
          self.cache.put(
              self,
              prompt,
              result.clone(override=dict(is_cached=True)),
              seed=cache_seed
          )
    return results  # pytype: disable=bad-return-type

  @abc.abstractmethod
  def _sample(
      self,
      prompt: list[message_lib.Message],
  ) -> list[LMSamplingResult]:
    """Subclass should override."""

  def _parallel_execute_with_currency_control(
      self,
      action: Callable[..., LMSamplingResult],
      inputs: Sequence[Any],
      retry_on_errors: Union[
          None,
          Union[Type[BaseException], Tuple[Type[BaseException], str]],
          Sequence[Union[Type[BaseException], Tuple[Type[BaseException], str]]],
      ] = RetryableLMError,
  ) -> list[Any]:
    """Helper method for subclasses for implementing _sample."""
    if self.max_concurrency is None:
      execute = action
      executor = None
      max_workers = len(inputs)
    else:
      execute = lambda x: _ConcurrencyControl.get(
          self.resource_id, self.max_concurrency)(action, x)
      executor = self.resource_id if len(inputs) > 1 else None
      max_workers = self.max_concurrency

    executed_jobs = concurrent.concurrent_execute(
        execute,
        inputs,
        executor=executor,
        max_workers=max_workers,
        retry_on_errors=retry_on_errors,
        max_attempts=self.max_attempts,
        retry_interval=self.retry_interval,
        exponential_backoff=self.exponential_backoff,
        max_retry_interval=self.max_retry_interval,
        return_jobs=True,
    )
    for job in executed_jobs:
      if isinstance(job.result, LMSamplingResult):
        job.result.usage.rebind(
            retry_stats=RetryStats.from_retry_entries(job.retry_entries),
            skip_notification=True,
        )
    return [job.result for job in executed_jobs]

  def __call__(
      self,
      prompt: str | message_lib.Message,
      *,
      cache_seed: int = 0, **kwargs
  ) -> message_lib.Message:
    """Returns the first candidate."""
    prompt = message_lib.UserMessage.from_value(prompt)
    with component.context(override_attrs=True, **kwargs):
      sampling_options = self.sampling_options
      if sampling_options.n != 1:
        sampling_options = sampling_options.clone(override=dict(n=1))

      call_counter = self._call_counter
      self._call_counter += 1
      request_start = time.time()
      result = self.sample(
          [prompt], sampling_options=sampling_options, cache_seed=cache_seed
      )[0]
      elapse = time.time() - request_start
      response = result.samples[0].response
      self._debug(prompt, response, call_counter, result.usage, elapse)
      return response

  def _debug(
      self,
      prompt: message_lib.Message,
      response: message_lib.Message,
      call_counter: int,
      usage: LMSamplingUsage,
      elapse: float,
  ) -> None:
    """Outputs debugging information."""
    debug = self.debug
    if isinstance(debug, bool):
      debug = LMDebugMode.ALL if debug else LMDebugMode.NONE

    if debug & LMDebugMode.INFO:
      self._debug_model_info(call_counter, usage)

    if debug & LMDebugMode.PROMPT:
      self._debug_prompt(prompt, call_counter, usage)

    if debug & LMDebugMode.RESPONSE:
      self._debug_response(response, call_counter, usage, elapse)

  def _debug_model_info(
      self, call_counter: int, usage: LMSamplingUsage) -> None:
    """Outputs debugging information about the model."""
    title_suffix = ''
    if usage.total_tokens != 0:
      title_suffix = pg.colored(
          f' (total {usage.total_tokens} tokens)', 'red'
      )

    console.write(
        self.format(compact=True, use_inferred=True),
        title=f'[{call_counter}] LM INFO{title_suffix}:',
        color='magenta',
    )

  def _debug_prompt(
      self,
      prompt: message_lib.Message,
      call_counter: int,
      usage: LMSamplingUsage,
  ) -> None:
    """Outputs debugging information about the prompt."""
    title_suffix = ''
    if usage.prompt_tokens != 0:
      title_suffix = pg.colored(f' ({usage.prompt_tokens} tokens)', 'red')

    console.write(
        # We use metadata 'formatted_text' for scenarios where the prompt text
        # is formatted by the LM.
        prompt.get('formatted_text', prompt.text),
        title=f'\n[{call_counter}] PROMPT SENT TO LM{title_suffix}:',
        color='green',
    )
    referred_modalities = prompt.referred_modalities()
    if referred_modalities:
      console.write(
          pg.object_utils.kvlist_str(
              [(k, repr(v), None) for k, v in referred_modalities.items()]
          ),
          title=f'\n[{call_counter}] MODALITY OBJECTS SENT TO LM:',
          color='green',
      )

  def _debug_response(
      self,
      response: message_lib.Message,
      call_counter: int,
      usage: LMSamplingUsage,
      elapse: float
  ) -> None:
    """Outputs debugging information about the response."""
    title_suffix = ' ('
    if usage.completion_tokens != 0:
      title_suffix += f'{usage.completion_tokens} tokens '
    title_suffix += f'in {elapse:.2f} seconds)'
    title_suffix = pg.colored(title_suffix, 'red')

    console.write(
        str(response) + '\n',
        title=f'\n[{call_counter}] LM RESPONSE{title_suffix}:',
        color='blue',
    )

  def score(
      self,
      prompt: str | message_lib.Message | list[message_lib.Message],
      completions: list[str | message_lib.Message],
      **kwargs,
  ) -> list[LMScoringResult]:
    """Scores the given prompt."""
    if isinstance(prompt, list):
      if len(prompt) != len(completions):
        raise ValueError(
            'prompt and completions must have the same length.'
        )
      prompt = [message_lib.UserMessage.from_value(p) for p in prompt]
    else:
      prompt = message_lib.UserMessage.from_value(prompt)
    completions = [message_lib.UserMessage.from_value(c) for c in completions]

    call_counter = self._call_counter
    self._call_counter += 1
    request_start = time.time()

    with component.context(override_attrs=True, **kwargs):
      scoring_results = self._score(prompt, completions)
      elapse = time.time() - request_start
      self._debug_score(
          prompt, completions, scoring_results, call_counter, elapse
      )
      return scoring_results

  def _score(
      self, prompt: message_lib.Message | list[message_lib.Message],
      completions: list[message_lib.Message]
  ) -> list[LMScoringResult]:
    """Subclass to implement."""
    raise NotImplementedError(
        f'{self.__class__.__name__} does not support scoring.'
    )

  def _debug_score(
      self,
      prompt: message_lib.Message | list[message_lib.Message],
      completions: list[message_lib.Message],
      scoring_results: list[LMScoringResult],
      call_counter: int,
      elapse: float,
  ):
    debug = self.debug
    if isinstance(debug, bool):
      debug = LMDebugMode.ALL if debug else LMDebugMode.NONE

    if debug & LMDebugMode.INFO:
      self._debug_model_info(call_counter, UsageNotAvailable())

    if debug & LMDebugMode.PROMPT:
      console.write(
          prompt,
          title=f'\n[{call_counter}] SCORING LM WITH PROMPT:',
          color='green',
      )
      if isinstance(prompt, list):
        referred_modalities_lst = [p.referred_modalities() for p in prompt]
      else:
        referred_modalities_lst = [prompt.referred_modalities(),]
      if referred_modalities_lst:
        for referred_modalities in referred_modalities_lst:
          console.write(
              pg.object_utils.kvlist_str(
                  [(k, repr(v), None) for k, v in referred_modalities.items()]
              ),
              title=f'\n[{call_counter}] MODALITY OBJECTS SENT TO LM:',
              color='green',
          )

    if debug & LMDebugMode.RESPONSE:
      console.write(
          '',
          title=(
              f'\n[{call_counter}] SCORING COMPLETED (in {elapse:.2f} seconds):'
          ),
          color='blue',
      )
      for i, (c, r) in enumerate(zip(completions, scoring_results)):
        console.write(
            c,
            title=f'COMPLETION #{i}',
            color='green',
        )
        console.write(
            f'score: {r.score}',
            color='blue',
        )

  def tokenize(
      self,
      prompt: str | message_lib.Message,
      **kwargs,
  ) -> list[tuple[str | bytes, int]]:
    """Tokenizes the given prompt."""
    prompt = message_lib.UserMessage.from_value(prompt)
    call_counter = self._call_counter
    self._call_counter += 1

    with component.context(override_attrs=True, **kwargs):
      request_start = time.time()
      tokens = self._tokenize(prompt)
      elapse = time.time() - request_start
      self._debug_tokenize(prompt, tokens, call_counter, elapse)
      return tokens

  def _tokenize(
      self, prompt: message_lib.Message
  ) -> list[tuple[str | bytes, int]]:
    """Subclass to implement."""
    raise NotImplementedError(
        f'{self.__class__.__name__} does not support tokenization.'
    )

  def _debug_tokenize(
      self,
      prompt: message_lib.Message,
      tokens: list[tuple[str | bytes, int]],
      call_counter: int,
      elapse: float,
  ):
    debug = self.debug
    if isinstance(debug, bool):
      debug = LMDebugMode.ALL if debug else LMDebugMode.NONE

    if debug & LMDebugMode.INFO:
      self._debug_model_info(call_counter, UsageNotAvailable())

    if debug & LMDebugMode.PROMPT:
      console.write(
          prompt,
          title=f'\n[{call_counter}] PROMPT TO TOKENIZE:',
          color='green',
      )
      referred_modalities_lst = [prompt.referred_modalities(),]
      if referred_modalities_lst:
        for referred_modalities in referred_modalities_lst:
          console.write(
              pg.object_utils.kvlist_str(
                  [(k, repr(v), None) for k, v in referred_modalities.items()]
              ),
              title=f'\n[{call_counter}] MODALITY OBJECTS SENT TO LM:',
              color='green',
          )

    if debug & LMDebugMode.RESPONSE:
      console.write(
          tokens,
          title=(
              f'\n[{call_counter}] {len(tokens)} TOKENS RETURNED '
              f'(in {elapse:.2f} seconds):'
          ),
          color='blue',
      )

  @classmethod
  def estimate_max_concurrency(
      cls,
      max_tokens_per_minute: int | None,
      max_requests_per_minute: int | None,
      average_tokens_per_request: int = 250
  ) -> int | None:
    """Estimates max concurrency concurrency based on the rate limits."""
    # NOTE(daiyip): max concurrency is estimated based on the rate limit.
    # We assume each request has approximately 250 tokens, and each request
    # takes 1 second to complete. This might not be accurate for all models.
    if max_tokens_per_minute is not None:
      return max(
          int(max_tokens_per_minute / average_tokens_per_request / 60), 1
      )
    elif max_requests_per_minute is not None:
      return max(int(max_requests_per_minute / 60), 1)
    return None


class _ConcurrencyControl:
  """Controls the max concurrent LLM calls for a given model."""

  _MODEL_CONCURRENCY: ClassVar[dict[str, '_ConcurrencyControl']] = {}

  def __init__(self, max_concurrency: int):
    self.max_concurrency = max_concurrency
    self._concurrency = 0

  @property
  def concurrency(self) -> int:
    """Returns the current concurrency."""
    return self._concurrency

  def __call__(self, fn: Callable[..., Any], *args, **kwargs):
    """Calls the function with concurrency control."""
    while self._concurrency >= self.max_concurrency:
      time.sleep(0.01)

    try:
      # Increment/decrement is atomic in Python, so we don't need to protect it
      # with a lock.
      self._concurrency += 1
      return fn(*args, **kwargs)
    finally:
      self._concurrency -= 1

  @classmethod
  def get(
      cls, model_id: str, max_concurrency: int | None = None
  ) -> '_ConcurrencyControl':
    """Returns the concurrency control for the given model ID."""
    control = cls._MODEL_CONCURRENCY.get(model_id, None)
    if control is None:
      assert max_concurrency is not None
      control = cls(max_concurrency)
      cls._MODEL_CONCURRENCY[model_id] = control
    return control


class UsageSummary(pg.Object, pg.views.HtmlTreeView.Extension):
  """Usage sumary."""

  class AggregatedUsage(pg.Object):
    """Aggregated usage."""

    total: LMSamplingUsage = LMSamplingUsage(0, 0, 0, 0, 0.0)
    breakdown: dict[str, LMSamplingUsage] = {}

    def __bool__(self) -> bool:
      """Returns True if the usage is non-empty."""
      return bool(self.breakdown)

    def add(
        self,
        model_id: str,
        usage: LMSamplingUsage,
    ) -> None:
      """Adds an entry to the breakdown."""
      aggregated = self.breakdown.get(model_id, None)
      with pg.notify_on_change(False):
        self.breakdown[model_id] = usage + aggregated
        self.rebind(
            total=self.total + usage,
            raise_on_no_change=False
        )

    def merge(self, other: 'UsageSummary.AggregatedUsage') -> None:
      """Merges the usage summary."""
      with pg.notify_on_change(False):
        for model_id, usage in other.breakdown.items():
          self.add(model_id, usage)

  def _on_bound(self):
    super()._on_bound()
    self._usage_badge = None
    self._lock = threading.Lock()

  @property
  def total(self) -> LMSamplingUsage:
    return self.cached.total + self.uncached.total

  def add(self, model_id: str, usage: LMSamplingUsage, is_cached: bool):
    """Updates the usage summary."""
    with self._lock:
      if is_cached:
        usage.rebind(estimated_cost=0.0, skip_notification=True)
        self.cached.add(model_id, usage)
      else:
        self.uncached.add(model_id, usage)
      self._update_view()

  def merge(self, other: 'UsageSummary', as_cached: bool = False) -> None:
    """Aggregates the usage summary.

    Args:
      other: The usage summary to merge.
      as_cached: Whether to merge the usage summary as cached.
    """
    with self._lock:
      self.cached.merge(other.cached)
      if as_cached:
        self.cached.merge(other.uncached)
      else:
        self.uncached.merge(other.uncached)
      self._update_view()

  def _sym_nondefault(self) -> dict[str, Any]:
    """Overrides nondefault values so volatile values are not included."""
    return dict()

  #
  # Html views for the usage summary.
  #

  def _update_view(self):
    if self._usage_badge is not None:
      self._usage_badge.update(
          self._badge_text(),
          tooltip=pg.format(
              self, verbose=False, custom_format=self._tooltip_format,
              hide_default_values=True,
          ),
          styles=dict(color=self._badge_color()),
      )

  def _badge_text(self) -> str:
    if self.total.estimated_cost is not None:
      return f'{self.total.estimated_cost:.3f}'
    return '0.000'

  def _badge_color(self) -> str | None:
    if self.total.estimated_cost is None or self.total.estimated_cost < 1.0:
      return None

    # Step 1: The normal cost range is around 1e-3 to 1e5.
    # Therefore we normalize the log10 value from [-3, 5] to [0, 1].
    normalized_value = (math.log10(self.total.estimated_cost) + 3) / (5 + 3)

    # Step 2: Interpolate between green and red
    red = int(255 * normalized_value)
    green = int(255 * (1 - normalized_value))
    return f'rgb({red}, {green}, 0)'

  def _tooltip_format(self, v, root_indent):
    del root_indent
    if isinstance(v, int):
      return f'{v:,}'
    if isinstance(v, float):
      return f'{v:,.3f}'
    return None

  def _html_tree_view(
      self,
      *,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ) -> pg.Html:
    extra_flags = extra_flags or {}
    as_badge = extra_flags.pop('as_badge', False)
    interactive = extra_flags.get('interactive', True)
    if as_badge:
      usage_badge = self._usage_badge
      if usage_badge is None:
        usage_badge = pg.views.html.controls.Badge(
            self._badge_text(),
            tooltip=pg.format(
                self, custom_format=self._tooltip_format, verbose=False,
                hide_default_values=True,
            ),
            css_classes=['usage-summary'],
            styles=dict(color=self._badge_color()),
            interactive=True,
        )
        if interactive:
          self._usage_badge = usage_badge
      return usage_badge.to_html()
    return super()._html_tree_view(
        view=view,
        extra_flags=extra_flags,
        **kwargs
    )

  @classmethod
  @functools.cache
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .usage-summary.label {
            display: inline-flex;
            border-radius: 5px;
            padding: 5px;
            background-color: #f1f1f1;
            color: #CCC;
        }
        .usage-summary.label::before {
            content: '$';
        }
        """
    ]

pg.members(
    dict(
        cached=(
            pg.typing.Object(
                UsageSummary.AggregatedUsage,
                default=UsageSummary.AggregatedUsage()
            ),
            'Aggregated usages for cached LLM calls.'
        ),
        uncached=(
            pg.typing.Object(
                UsageSummary.AggregatedUsage,
                default=UsageSummary.AggregatedUsage()
            ),
            'Aggregated usages for uncached LLM calls.'
        ),
    )
)(UsageSummary)


class _UsageTracker:
  """Usage tracker."""

  def __init__(self, model_ids: set[str] | None):
    self.model_ids = model_ids
    self.usage_summary = UsageSummary()

  def track(self, model_id: str, usage: LMSamplingUsage, is_cached: bool):
    if self.model_ids is None or model_id in self.model_ids:
      self.usage_summary.add(model_id, usage, is_cached)


@contextlib.contextmanager
def track_usages(
    *lm: Union[str, LanguageModel]
) -> Iterator[UsageSummary]:
  """Context manager to track the usages of all language models in scope.

  `lf.track_usages` works with threads spawned by `lf.concurrent_map` and
  `lf.concurrent_execute`.

  Example:
    ```
    lm = lf.llms.GeminiPro1()
    with lf.track_usages() as usages:
      # invoke any code that will call LLMs.

    print(usages.uncached.breakdown[lm.model_id])
    ```

  Args:
    *lm: The language model(s) to track. If None, track all models in scope.

  Yields:
    A dictionary of model ID to usage. If a model does not supports usage
    counting, the dict entry will be None.
  """
  if not lm:
    model_ids = None
  else:
    model_ids = [m.model_id if isinstance(m, LanguageModel) else m for m in lm]

  trackers = component.context_value('__usage_trackers__', [])
  tracker = _UsageTracker(set(model_ids) if model_ids else None)
  with component.context(__usage_trackers__=trackers + [tracker]):
    try:
      yield tracker.usage_summary
    finally:
      pass



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.console."""

import contextlib
import io
import unittest

from langfun.core import console
import pyglove as pg


class ConsoleTest(unittest.TestCase):

  def test_write(self):
    s = io.StringIO()
    with contextlib.redirect_stdout(s):
      console.write('foo', title='hello')
    s = s.getvalue()
    self.assertIn('hello', s)
    self.assertIn('foo', s)

  def test_under_notebook(self):
    self.assertFalse(console.under_notebook())
    console._notebook = True
    self.assertTrue(console.under_notebook())
    console._notebook = None

  def test_notebook_interaction(self):
    console._notebook = pg.Dict(
        display=lambda x: x, Javascript=lambda x: x, clear_output=lambda: None)
    self.assertEqual(console.display('hi', clear=True), 'hi')
    self.assertEqual(
        console.run_script('console.log("hi")'),
        'console.log("hi")'
    )
    console.clear()
    console._notebook = None
    self.assertIsNone(console.display('hi'))
    self.assertIsNone(console.run_script('console.log("hi")'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Langfun event logging."""

import contextlib
import datetime
import functools
import typing
from typing import Any, Callable, Iterator, Literal

from langfun.core import component
from langfun.core import console as console_lib
import pyglove as pg


LogLevel = Literal['debug', 'info', 'error', 'warning', 'fatal']
_LOG_LEVELS = list(typing.get_args(LogLevel))


@contextlib.contextmanager
def use_log_level(log_level: LogLevel = 'info') -> Iterator[None]:
  """Contextmanager to enable logging at a given level."""
  with component.context(__event_log_level__=log_level):
    try:
      yield
    finally:
      pass


def get_log_level() -> LogLevel:
  """Gets the current minimum log level."""
  return component.context_value('__event_log_level__', 'info')


def system_log_func(level: LogLevel) -> Callable[..., None]:
  match level:
    case 'debug':
      return pg.logging.debug
    case 'info':
      return pg.logging.info
    case 'warning':
      return pg.logging.warning
    case 'error':
      return pg.logging.error
    case 'fatal':
      return pg.logging.error
    case _:
      raise ValueError(f'Unsupported log level: {level}')


class LogEntry(pg.Object, pg.views.HtmlTreeView.Extension):
  """Event log entry."""
  time: datetime.datetime
  level: LogLevel
  message: str
  metadata: dict[str, Any] = pg.Dict()
  indent: int = 0

  def should_output(self, min_log_level: LogLevel) -> bool:
    return _LOG_LEVELS.index(self.level) >= _LOG_LEVELS.index(min_log_level)

  def format(self,
             compact: bool = False,
             verbose: bool = True,
             root_indent: int = 0,
             *,
             text_format: bool = True,
             **kwargs):
    if text_format:
      s = f"""{self.time.strftime('%H:%M:%S')} {self.level.upper()} - {self.message}"""
      if self.metadata:
        s += f' (metadata: {self.metadata!r})'
      return s
    return super().format(
        compact=compact,
        verbose=verbose,
        root_indent=root_indent,
        **kwargs
    )

  def _html_tree_view_summary(
      self,
      view: pg.views.HtmlTreeView,
      title: str | pg.Html | None = None,
      max_summary_len_for_str: int = 80,
      **kwargs
  ) -> pg.Html | None:
    if len(self.message) > max_summary_len_for_str:
      message = self.message[:max_summary_len_for_str] + '...'
    else:
      message = self.message

    s = pg.Html(
        pg.Html.element(
            'span',
            [self.time.strftime('%H:%M:%S')],
            css_classes=['log-time']
        ),
        pg.Html.element(
            'span',
            [pg.Html.escape(message)],
            css_classes=['log-summary'],
        ),
    )
    return view.summary(
        self,
        title=title or s,
        max_summary_len_for_str=max_summary_len_for_str,
        **kwargs,
    )

  # pytype: disable=annotation-type-mismatch
  def _html_tree_view_content(
      self,
      view: pg.views.HtmlTreeView,
      root_path: pg.KeyPath | None = None,
      max_summary_len_for_str: int = 80,
      collapse_level: int | None = 1,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ) -> pg.Html:
    # pytype: enable=annotation-type-mismatch
    extra_flags = extra_flags if extra_flags is not None else {}
    collapse_log_metadata_level: int | None = extra_flags.get(
        'collapse_log_metadata_level', None
    )
    def render_message_text():
      if len(self.message) < max_summary_len_for_str:
        return None
      return pg.Html.element(
          'span',
          [pg.Html.escape(self.message)],
          css_classes=['log-text'],
      )

    def render_metadata():
      if not self.metadata:
        return None
      return pg.Html.element(
          'div',
          [
              view.render(
                  self.metadata,
                  name='metadata',
                  root_path=pg.KeyPath('metadata', root_path),
                  parent=self,
                  collapse_level=view.get_collapse_level(
                      (collapse_level, -1), collapse_log_metadata_level,
                  ),
                  max_summary_len_for_str=max_summary_len_for_str,
                  extra_flags=extra_flags,
                  **view.get_passthrough_kwargs(**kwargs),
              )
          ],
          css_classes=['log-metadata'],
      )

    return pg.Html.element(
        'div',
        [
            render_message_text(),
            render_metadata(),
        ],
        css_classes=['complex_value'],
    )

  def _html_tree_view_config(self) -> dict[str, Any]:
    return pg.views.HtmlTreeView.get_kwargs(
        super()._html_tree_view_config(),
        dict(
            css_classes=[f'log-{self.level}'],
        )
    )

  @classmethod
  @functools.cache
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        /* Langfun LogEntry styles. */
        .log-time {
          color: #222;
          font-size: 12px;
          padding-right: 10px;
        }
        .log-summary {
          font-weight: normal;
          font-style: italic;
          padding: 4px;
        }
        .log-debug > summary > .summary_title::before {
          content: '🛠️ '
        }
        .log-info > summary > .summary_title::before {
          content: '💡 '
        }
        .log-warning > summary > .summary_title::before {
          content: '❗ '
        }
        .log-error > summary > .summary_title::before {
          content: '❌ '
        }
        .log-fatal > summary > .summary_title::before {
          content: '💀 '
        }
        .log-text {
          display: block;
          color: black;
          font-style: italic;
          padding: 20px;
          border-radius: 5px;
          background: rgba(255, 255, 255, 0.5);
          white-space: pre-wrap;
        }
        details.log-entry {
          margin: 0px 0px 10px;
          border: 0px;
        }
        div.log-metadata {
          margin: 10px 0px 0px 0px; 
        }
        .log-metadata > details {
          background-color: rgba(255, 255, 255, 0.5);
          border: 1px solid transparent;
        }
        .log-debug {
          background-color: #EEEEEE
        }
        .log-warning {
          background-color: #F8C471
        }
        .log-info {
          background-color: #A3E4D7
        }
        .log-error {
          background-color: #F5C6CB
        }
        .log-fatal {
          background-color: #F19CBB
        }
        """
    ]


def log(level: LogLevel,
        message: str,
        *,
        indent: int = 0,
        console: bool = False,
        **kwargs) -> LogEntry:
  """Logs a message."""
  entry = LogEntry(
      indent=indent,
      level=level,
      time=datetime.datetime.now(),
      message=message,
      metadata=kwargs,
  )

  if console and entry.should_output(get_log_level()):
    if console_lib.under_notebook():
      console_lib.display(entry)
    else:
      # TODO(daiyip): Improve the console output formatting.
      console_lib.write(entry)
  elif not console:
    if kwargs:
      message = f'{message} (metadata: {pg.format(kwargs, verbose=False)})'
    system_log_func(level)(message)
  return entry


def debug(
    message: str,
    *,
    indent: int = 0,
    console: bool = False,
    **kwargs
) -> LogEntry:
  """Logs a debug message to the session."""
  return log('debug', message, indent=indent, console=console, **kwargs)


def info(
    message: str,
    *,
    indent: int = 0,
    console: bool = False,
    **kwargs
) -> LogEntry:
  """Logs an info message to the session."""
  return log('info', message, indent=indent, console=console, **kwargs)


def warning(
    message: str,
    *,
    indent: int = 0,
    console: bool = False,
    **kwargs
) -> LogEntry:
  """Logs an info message to the session."""
  return log('warning', message, indent=indent, console=console, **kwargs)


def error(
    message: str,
    *,
    indent: int = 0,
    console: bool = False,
    **kwargs
) -> LogEntry:
  """Logs an error message to the session."""
  return log('error', message, indent=indent, console=console, **kwargs)


def fatal(
    message: str,
    *,
    indent: int = 0,
    console: bool = False,
    **kwargs
) -> LogEntry:
  """Logs a fatal message to the session."""
  return log('fatal', message, indent=indent, console=console, **kwargs)


# Register the logging functions to skip the source of the logging functions.
pg.logging.register_frame_to_skip([log, debug, info, warning, error, fatal])



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Progress reporting for evaluation."""

import datetime
import threading
import time
from typing import Annotated, Any
import pyglove as pg


class Progress(pg.Object, pg.views.HtmlTreeView.Extension):
  """Evaluation progress."""

  num_total: Annotated[
      int | None,
      (
          'Total number of items to be processed. '
          'If None, the progress is not started.'
      )
  ] = None
  num_processed: Annotated[
      int,
      (
          'Number of items that have been processed without errors.'
      )
  ] = 0
  num_failed: Annotated[
      int,
      (
          'Number of items that have failed.'
      )
  ] = 0
  num_skipped: Annotated[
      int,
      (
          'Number of items that have been skipped.'
      )
  ] = 0
  start_time: Annotated[
      float | None,
      (
          'The start time of the progress. '
          'If None, the progress is not started.'
      )
  ] = None
  stop_time: Annotated[
      float | None,
      (
          'The stop time of the progress. '
          'If None, the progress is not stopped.'
      )
  ] = None
  execution_summary: Annotated[
      pg.object_utils.TimeIt.StatusSummary,
      'The execution summary of the progress.'
  ] = pg.object_utils.TimeIt.StatusSummary()

  def _on_bound(self):
    super()._on_bound()
    self._progress_bar = None
    self._time_label = None
    self._lock = threading.Lock()

  def reset(self) -> None:
    """Resets the progress."""
    self._sync_members(
        num_total=None,
        num_processed=0,
        num_failed=0,
        num_skipped=0,
        start_time=None,
        stop_time=None,
        execution_summary=pg.object_utils.TimeIt.StatusSummary(),
    )

  @property
  def num_completed(self) -> int:
    """Returns the number of completed examples."""
    return self.num_processed + self.num_failed + self.num_skipped

  def __float__(self) -> float:
    """Returns the complete rate in range [0, 1]."""
    if self.num_total is None:
      return float('nan')
    return self.num_completed / self.num_total

  @property
  def is_started(self) -> bool:
    """Returns whether the evaluation is started."""
    return self.start_time is not None

  @property
  def is_stopped(self) -> bool:
    """Returns whether the evaluation is stopped."""
    return self.stop_time is not None

  @property
  def is_completed(self) -> bool:
    """Returns whether the evaluation is completed."""
    return (
        self.num_total is not None
        and self.num_completed == self.num_total
    )

  @property
  def is_skipped(self) -> bool:
    """Returns whether the evaluation is skipped."""
    return (
        self.num_total is not None
        and self.num_skipped == self.num_total
    )

  @property
  def is_failed(self) -> bool:
    """Returns whether the evaluation is failed."""
    return (
        self.num_failed > 0
        and self.num_failed + self.num_skipped == self.num_total
    )

  @property
  def elapse(self) -> float | None:
    """Returns the elapse time in seconds."""
    if self.start_time is None:
      return None
    if self.stop_time is None:
      return time.time() - self.start_time
    return self.stop_time - self.start_time

  @property
  def start_time_str(self) -> str | None:
    """Returns the start time string of the evaluation."""
    if self.start_time is None:
      return None
    return time.strftime('%Y/%m/%d %H:%M:%S', time.localtime(self.start_time))

  @property
  def stop_time_str(self) -> str | None:
    """Returns the complete time string of the evaluation."""
    if self.stop_time is None:
      return None
    return time.strftime(
        '%Y/%m/%d %H:%M:%S', time.localtime(self.stop_time)
    )

  def start(self, total: int) -> None:
    """Marks the evaluation as started."""
    assert self.start_time is None, self
    self._sync_members(start_time=time.time(), num_total=total)
    if self._progress_bar is not None:
      self._progress_bar.update(total=total)
    self._update_time_label()

  def stop(self) -> None:
    """Marks the evaluation as stopped."""
    assert self.stop_time is None, self
    self._sync_members(stop_time=time.time())
    self._update_time_label()

  def _sync_members(self, **kwargs: Any):
    """Synchronizes the members of the progress."""
    self.rebind(
        **kwargs,
        skip_notification=True,
        raise_on_no_change=False,
    )

  def increment_processed(self, delta: int = 1) -> None:
    """Updates the number of processed examples."""
    assert self.is_started and not self.is_stopped, self
    with self._lock:
      self._sync_members(num_processed=self.num_processed + delta)
    if self._progress_bar is not None:
      self._progress_bar['Processed'].increment(delta)
    self._update_time_label()

  def increment_failed(self, delta: int = 1) -> None:
    """Updates the number of failed examples."""
    assert self.is_started and not self.is_stopped, self
    with self._lock:
      self._sync_members(num_failed=self.num_failed + delta)
    if self._progress_bar is not None:
      self._progress_bar['Failed'].increment(delta)
    self._update_time_label()

  def increment_skipped(self, delta: int = 1) -> None:
    """Updates the number of skipped examples."""
    assert self.is_started and not self.is_stopped, self
    with self._lock:
      self._sync_members(num_skipped=self.num_skipped + delta)
    if self._progress_bar is not None:
      self._progress_bar['Skipped'].increment(delta)
    self._update_time_label()

  def update_execution_summary(
      self,
      execution_status: dict[str, pg.object_utils.TimeIt.Status]
  ) -> None:
    """Updates the execution summary of the progress."""
    with self._lock:
      self.execution_summary.aggregate(execution_status)

  def _sym_nondefault(self) -> dict[str, Any]:
    """Overrides nondefault values so volatile values are not included."""
    return dict()

  #
  # HTML view.
  #

  def _duration_text(self) -> str:
    if self.start_time is None:
      return '00:00:00'
    return str(datetime.timedelta(seconds=self.elapse)).split('.')[0]

  def _time_tooltip(self) -> pg.Html.WritableTypes:
    time_info = pg.Dict(
        duration=self._duration_text(),
        last_update=(
            time.strftime(    # pylint: disable=g-long-ternary
                '%Y/%m/%d %H:%M:%S',
                time.localtime(time.time())
            ) if not self.is_stopped else self.stop_time_str
        ),
        start_time=self.start_time_str,
        stop_time=self.stop_time_str,
    )
    if self.execution_summary:
      time_info['execution'] = pg.Dict(
          {
              k: pg.Dict(
                  num_started=v.num_started,
                  num_ended=v.num_ended,
                  num_failed=v.num_failed,
                  avg_duration=round(v.avg_duration, 2),
              ) for k, v in self.execution_summary.breakdown.items()
          }
      )
    return pg.format(time_info, verbose=False)

  def _html_tree_view(
      self,
      *,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ) -> pg.Html:
    """Renders the content of the progress bar."""
    def _progress_bar():
      return pg.views.html.controls.ProgressBar(
          [
              pg.views.html.controls.SubProgress(
                  name='Skipped', value=self.num_skipped,
              ),
              pg.views.html.controls.SubProgress(
                  name='Processed', value=self.num_processed,
              ),
              pg.views.html.controls.SubProgress(
                  name='Failed', value=self.num_failed,
              ),
          ],
          total=self.num_total,
          interactive=interactive,
      )

    def _time_label():
      css_class = 'not-started'
      if self.is_started and not self.is_stopped:
        css_class = 'started'
      elif self.is_stopped:
        css_class = 'stopped'
      return pg.views.html.controls.Label(
          self._duration_text(),
          tooltip=self._time_tooltip(),
          css_classes=[
              'progress-time', css_class
          ],
          interactive=interactive,
      )

    extra_flags = extra_flags or {}
    interactive = extra_flags.pop('interactive', True)
    if interactive:
      if self._progress_bar is None:
        self._progress_bar = _progress_bar()
      if self._time_label is None:
        self._time_label = _time_label()
      progress_bar = self._progress_bar
      time_label = self._time_label
    else:
      progress_bar = _progress_bar()
      time_label = _time_label()
    return pg.Html.element(
        'div', [progress_bar, time_label], css_classes=['eval-progress'],
    )

  def _update_time_label(self):
    """Updates the time label of the progress."""
    if self._time_label is None:
      return
    self._time_label.update(
        text=self._duration_text(),
        tooltip=self._time_tooltip(),
        styles=dict(
            color=(
                'dodgerblue' if self.is_started
                and not self.is_stopped else '#ccc'
            ),
        ),
    )

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .eval-progress {
          display: inline-block;
        }
        .sub-progress.skipped {
          background-color:yellow;
        }
        .sub-progress.processed {
          background-color:#00B000;
        }
        .sub-progress.failed {
          background-color:red;
        }
        .progress-time {
          font-weight: normal;
          margin-left: 10px;
          border-radius: 5px;
          color: #CCC;
          padding: 5px;
        }
        """
    ]



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic completion."""

from typing import Annotated, Any, Type

import langfun.core as lf
from langfun.core.structured import mapping
from langfun.core.structured import schema as schema_lib
import pyglove as pg


class _CompleteStructure(mapping.Mapping):
  """Complete structure by filling the missing fields."""

  input: Annotated[
      pg.Symbolic, 'A symbolic object with `lf.MISSING` values to complete.'
  ] = lf.contextual()

  mapping_template = lf.Template("""
      {{ input_title }}:
      {{ example.input_repr(use_modality_ref=True) | indent(2, True) }}

      {%- if missing_type_dependencies(example.input) %}

      {{ schema_title }}:
      {{ class_defs_repr(example.input) | indent(2, True) }}
      {%- endif %}
      {%- if has_modality_refs(example.input) %}

      {{ modality_refs_title }}:
      {{ modality_refs_repr(example.input) | indent(2, True) }}
      {%- endif %}

      {{ output_title }}:
      {%- if example.has_output %}
      {{ example.output_repr(use_modality_ref=True) | indent(2, True) }}
      {% endif -%}
      """)

  input_title = 'INPUT_OBJECT'
  output_title = 'OUTPUT_OBJECT'
  schema_title = 'CLASS_DEFINITIONS'
  modality_refs_title: Annotated[
      str, 'The section title for modality refs.'
  ] = 'MODALITY_REFERENCES'

  preamble = lf.LangFunc(
      """
      Please generate the OUTPUT_OBJECT by completing the MISSING fields from the last INPUT_OBJECT.

      INSTRUCTIONS:
      1. Each MISSING field contains a Python annotation, please fill the value based on the annotation.
      2. Classes for the MISSING fields are defined under CLASS_DEFINITIONS.

      {{input_title}}:
        ```python
        Answer(
          question='1 + 1 =',
          answer=MISSING(int)
        )
        ```

      {{output_title}}:
        ```python
        Answer(
          question='1 + 1 =',
          answer=2
        )
        ```
      """,
      input_title=input_title,
      output_title=output_title,
  )

  @property
  def mapping_request(self) -> mapping.MappingExample:
    """Returns a MappingExample as the mapping request."""
    return mapping.MappingExample(
        input=pg.Ref(self.input),
        # Use the schema of input object.
        schema=pg.Ref(schema_lib.Schema.from_value(self.input.__class__)),
        context=self.context,
    )

  def transform_input(self, lm_input: lf.Message) -> lf.Message:
    if not pg.contains(self.input, type=schema_lib.Missing):
      raise ValueError(
          'The input of `lf.complete` must contain a least one '
          f'missing value. Encountered: {self.input}.'
      )
    return super().transform_input(lm_input)

  def missing_type_dependencies(self, value: Any) -> list[Type[Any]]:
    value_specs = tuple(
        [v.value_spec for v in schema_lib.Missing.find_missing(value).values()]
    )
    return schema_lib.class_dependencies(value_specs, include_subclasses=True)

  def class_defs_repr(self, value: Any) -> str | None:
    return schema_lib.class_definitions(
        self.missing_type_dependencies(value),
        markdown=True,
        allowed_dependencies=set()
    )

  def postprocess_result(self, result: Any) -> Any:
    """Postprocess result."""
    # Try restore modality objects from the input value to output value.
    modalities = self.modalities(self.input)
    if modalities:
      # Remove the `input` prefix for all entries.
      modalities = pg.object_utils.flatten(
          pg.object_utils.canonicalize(modalities)['input']
      )
      result.rebind(modalities)
    return result

  def globals(self):
    context = super().globals()

    # Add class dependencies from the input value to the globals.
    classes = schema_lib.class_dependencies(self.input)
    context.update({cls.__name__: cls for cls in classes})

    # NOTE(daiyip): since `lf.complete` could have fields of Any type, which
    # could be user provided objects. For LLMs to restores these objects, we
    # need to expose their types to the code evaluation context.
    context.update(self._input_value_dependencies())
    return context

  def _input_value_dependencies(self) -> dict[str, Any]:
    """Returns the class dependencies from input value."""
    context = {}
    def _visit(k, v, p):
      del k, p
      if isinstance(v, pg.Object):
        cls = v.__class__
        context[cls.__name__] = cls
    pg.traverse(self.input, _visit)
    return context

  #
  # Helper methods for handling modalities.
  #

  def has_modality_refs(self, value: Any) -> bool:
    """Returns true if the value has modalities."""
    return not isinstance(value, lf.Modality) and pg.contains(
        value, type=lf.Modality
    )

  def modalities(self, value: Any) -> dict[str, lf.Modality]:
    return lf.Modality.from_value(value)

  def modality_refs_repr(self, value: Any) -> str:
    with lf.modality.format_modality_as_ref(True):
      return pg.format(
          self.modalities(value),
          compact=False,
          verbose=False,
          python_format=True,
      )


def complete(
    input_value: pg.Symbolic,
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    cache_seed: int | None = 0,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    returns_message: bool = False,
    **kwargs,
) -> Any:
  """Complete a symbolic value by filling its missing fields.

  Examples:

    ```
    class FlightDuration:
      hours: int
      minutes: int

    class Flight(pg.Object):
      airline: str
      flight_number: str
      departure_airport_code: str
      arrival_airport_code: str
      departure_time: str
      arrival_time: str
      duration: FlightDuration
      stops: int
      price: float

    prompt = '''
      Information about flight UA2631.
      '''

    r = lf.query(prompt, Flight)
    assert isinstance(r, Flight)
    assert r.airline == 'United Airlines'
    assert r.departure_airport_code == 'SFO'
    assert r.duration.hour = 7
    ```

  Args:
    input_value: A symbolic value that may contain missing values.
    default: The default value if parsing failed. If not specified, error will
      be raised.
    lm: The language model to use. If not specified, the language model from
      `lf.context` context manager will be used.
    examples: An optional list of fewshot examples for helping parsing. If None,
      the default one-shot example will be added.
    cache_seed: Seed for computing cache key. The cache key is determined by a
      tuple of (lm, prompt, cache seed). If None, cache will be disabled for
      the query even cache is configured by the LM.
    autofix: Number of attempts to auto fix the generated code. If 0, autofix is
      disabled.
    autofix_lm: The language model to use for autofix. If not specified, the
      `autofix_lm` from `lf.context` context manager will be used. Otherwise it
      will use `lm`.
    returns_message: If True, returns `lf.Message` as the output, instead of
      returning the structured `message.result`.
    **kwargs: Keyword arguments passed to the
      `lf.structured.NaturalLanguageToStructureed` transform.

  Returns:
    The result based on the schema.
  """
  t = _CompleteStructure(
      input=schema_lib.mark_missing(input_value),
      default=default,
      examples=examples,
      autofix=autofix,
      **kwargs,
  )

  output = t(lm=lm, cache_seed=cache_seed, autofix_lm=autofix_lm or lm)
  return output if returns_message else output.result


async def acomplete(
    input_value: pg.Symbolic,
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    cache_seed: int | None = 0,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    returns_message: bool = False,
    **kwargs,
) -> Any:
  """Async version of `lf.complete`."""
  # TODO(daiyip): implement native async completion.
  return await lf.invoke_async(
      complete,
      input_value,
      default,
      lm=lm,
      examples=examples,
      cache_seed=cache_seed,
      autofix=autofix,
      autofix_lm=autofix_lm,
      returns_message=returns_message,
      **kwargs
  )



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for scoring evaluation."""

import os
import tempfile
import unittest

import langfun.core as lf
from langfun.core.eval import scoring
from langfun.core.llms import fake
import pyglove as pg


@pg.functor()
def float_list():
  return list[float]


@pg.functor()
def constrained_by_upperbound(upper_bound: int):
  return [
      (
          'Give an example of three numbers (x, y, z) '
          f'such that x + y + z <={upper_bound}',
      ),
      (
          'Give an example of two numbers (x, y) '
          f'such that x + y <={upper_bound}',
      ),
  ]


class ConstraintFollowing(scoring.Scoring):
  inputs = constrained_by_upperbound(1)
  prompt = '{{example}}'
  method = 'query'
  schema_fn = float_list()
  use_cache = True
  max_workers = 1

  def score(self, example, output):
    return 1.0 if sum(output) <= self.inputs.upper_bound else 0.0


def eval_set(lm: lf.LanguageModel):
  """Creates an evaluation object for testing."""
  tmp_dir = tempfile.gettempdir()
  return ConstraintFollowing(root_dir=tmp_dir, lm=lm)


class ScoringTest(unittest.TestCase):
  """Scoring test."""

  def setUp(self):
    super().setUp()
    pg.symbolic.set_save_handler(pg.symbolic.default_save_handler)
    pg.symbolic.set_load_handler(pg.symbolic.default_load_handler)

  def test_run(self):
    lm = fake.StaticSequence([
        '[0.5, 0.2, 0.3]',
        '[0.6, 0.7]',
    ])

    s = eval_set(lm=lm)
    self.assertEqual(s.avg_score, 0.0)
    s.run()
    result_copy = s.result.copy()
    del result_copy['experiment_setup']['id']
    self.assertEqual(
        result_copy,
        dict(
            experiment_setup=dict(
                dir=s.dir,
                model='StaticSequence',
                prompt_template='{{example}}',
                method='query',
                schema_fn='float_list()',
            ),
            cache_stats=dict(
                use_cache=True, num_queries=2, num_hits=0, num_updates=2
            ),
            metrics=dict(
                total=2,
                failures=0,
                failure_rate=0.0,
                oop_failures=0,
                oop_failure_rate=0.0,
                non_oop_failures=0,
                non_oop_failure_rate=0.0,
                failure_breakdown={},
                num_scored=2,
                score_rate=1.0,
                avg_score=0.5,
            ),
            usage=s.result.usage,
        ),
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, scoring.Scoring.EXPERIMENT_JSON
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, scoring.Scoring.CACHE_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, scoring.Scoring.RESULT_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, scoring.Scoring.OOP_FAILURES_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, scoring.Scoring.NON_OOP_FAILURES_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, scoring.Scoring.SCORED_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(os.path.join(s.root_dir, scoring.Scoring.SUMMARY_HTML))
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, scoring.Scoring.INDEX_HTML)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, scoring.Scoring.OOP_FAILURES_HTML
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, scoring.Scoring.NON_OOP_FAILURES_HTML
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, scoring.Scoring.SCORED_HTML
            )
        )
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structured parsing."""

import dataclasses
import inspect
import typing
import unittest

import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.structured import schema as schema_lib
import pyglove as pg


class Activity(pg.Object):
  description: str


class Itinerary(pg.Object):
  """A travel itinerary for a day."""

  day: pg.typing.Int[1, None]
  type: pg.typing.Enum['daytime', 'nighttime']
  activities: list[Activity]
  hotel: pg.typing.Annotated[
      pg.typing.Str['.*Hotel'] | None,
      'Hotel to stay if applicable.'
  ]


class PlaceOfInterest(pg.Object):
  """The name of a place of interest.

  A place of interest is a place that people ususally visit during their
  travels.
  """

  name: str


Itinerary.__serialization_key__ = 'Itinerary'


class Node(pg.Object):
  children: list['Node']


class SchemaTest(unittest.TestCase):

  def assert_schema(self, annotation, spec):
    self.assertEqual(schema_lib.Schema(annotation).spec, spec)

  def assert_unsupported_annotation(self, annotation):
    with self.assertRaises(ValueError):
      schema_lib.Schema(annotation)

  def test_init(self):
    self.assert_schema(int, pg.typing.Int())
    self.assert_schema(float, pg.typing.Float())
    self.assert_schema(str, pg.typing.Str())
    self.assert_schema(bool, pg.typing.Bool())

    # Top-level dictionary with 'result' as the only key is flattened.
    self.assert_schema(dict(result=int), pg.typing.Int())

    self.assert_schema(list[str], pg.typing.List(pg.typing.Str()))
    self.assert_schema([str], pg.typing.List(pg.typing.Str()))

    with self.assertRaisesRegex(
        ValueError, 'Annotation with list must be a list of a single element.'
    ):
      schema_lib.Schema([str, int])

    self.assert_schema(
        dict[str, int], pg.typing.Dict([(pg.typing.StrKey(), pg.typing.Int())])
    )

    self.assert_schema(
        {
            'x': int,
            'y': [str],
        },
        pg.typing.Dict([
            ('x', int),
            ('y', pg.typing.List(pg.typing.Str())),
        ]),
    )

    self.assert_schema(Itinerary, pg.typing.Object(Itinerary))

    self.assert_unsupported_annotation(typing.Type[int])
    self.assert_unsupported_annotation(typing.Union[int, str, bool])
    self.assert_unsupported_annotation(typing.Any)

  def test_schema_dict(self):
    schema = schema_lib.Schema([{'x': Itinerary}])
    self.assertEqual(
        schema.schema_dict(),
        {
            'result': [
                {
                    'x': {
                        '_type': 'Itinerary',
                        'day': pg.typing.Int(min_value=1),
                        'activities': [{
                            '_type': Activity.__type_name__,
                            'description': pg.typing.Str(),
                        }],
                        'hotel': pg.typing.Str(regex='.*Hotel').noneable(),
                        'type': pg.typing.Enum['daytime', 'nighttime'],
                    }
                }
            ]
        },
    )

  def test_class_dependencies(self):
    class Foo(pg.Object):
      x: int

    class Bar(pg.Object):
      y: str

    class A(pg.Object):
      foo: tuple[Foo, int]

    class X(pg.Object):
      k: int

    class B(A):
      bar: Bar
      foo2: Foo | X

    schema = schema_lib.Schema([B])
    v = schema_lib.structure_from_python(
        """
        class C(B):
          pass
        """,
        global_vars=dict(B=B),
        permission=pg.coding.CodePermission.ALL,
    )
    self.assertEqual(v.__module__, 'builtins')
    self.assertEqual(schema.class_dependencies(), [Foo, A, Bar, X, B])

  def test_class_dependencies_non_pyglove(self):
    class Baz:
      def __init__(self, x: int):
        pass

    @dataclasses.dataclass(frozen=True)
    class AA:
      foo: tuple[Baz, int]

    class XX(pg.Object):
      pass

    @dataclasses.dataclass(frozen=True)
    class BB(AA):
      foo2: Baz | XX

    v = schema_lib.structure_from_python(
        """
        class CC(BB):
          pass
        """,
        global_vars=dict(BB=BB),
        permission=pg.coding.CodePermission.ALL,
    )
    self.assertEqual(v.__module__, 'builtins')
    schema = schema_lib.Schema([AA])
    self.assertEqual(schema.class_dependencies(), [Baz, AA, XX, BB])

  def test_schema_repr(self):
    schema = schema_lib.Schema([{'x': Itinerary}])
    self.assertEqual(
        schema.schema_str(protocol='json'),
        (
            '{"result": [{"x": {"_type": "Itinerary", "day":'
            ' int(min=1), "type": "daytime" | "nighttime", "activities":'
            ' [{"_type": "%s", "description": str}], "hotel":'
            ' str(regex=.*Hotel) | None}}]}' % (
                Activity.__type_name__,
            )
        ),
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported protocol'):
      schema.schema_str(protocol='text')

  def test_value_repr(self):
    schema = schema_lib.Schema(int)
    self.assertEqual(schema.value_str(1, protocol='json'), '{"result": 1}')
    with self.assertRaisesRegex(ValueError, 'Unsupported protocol'):
      schema.value_str(1, protocol='text')

  def test_parse(self):
    schema = schema_lib.Schema(int)
    self.assertEqual(schema.parse('{"result": 1}'), 1)
    schema = schema_lib.Schema(dict[str, int])
    self.assertEqual(
        schema.parse('{"result": {"x": 1}}}'),
        dict(x=1)
    )
    with self.assertRaisesRegex(
        schema_lib.SchemaError, 'Expect .* but encountered .*'):
      schema.parse('{"result": "def"}')

    with self.assertRaisesRegex(ValueError, 'Unsupported protocol'):
      schema.parse('1', protocol='text')


class ClassDependenciesTest(unittest.TestCase):

  def test_class_dependencies_from_specs(self):
    class Foo(pg.Object):
      x: int

    class Bar(pg.Object):
      y: str

    class A(pg.Object):
      foo: tuple[Foo, int]

    class X(pg.Object):
      k: int

    class B(A):
      bar: Bar
      foo2: Foo | X

    self.assertEqual(schema_lib.class_dependencies(Foo), [Foo])

    self.assertEqual(
        schema_lib.class_dependencies((A,), include_subclasses=False), [Foo, A]
    )

    self.assertEqual(
        schema_lib.class_dependencies(A, include_subclasses=True),
        [Foo, A, Bar, X, B],
    )

    self.assertEqual(
        schema_lib.class_dependencies(schema_lib.Schema(A)), [Foo, A, Bar, X, B]
    )

    self.assertEqual(
        schema_lib.class_dependencies(pg.typing.Object(A)), [Foo, A, Bar, X, B]
    )

    with self.assertRaisesRegex(TypeError, 'Unsupported spec type'):
      schema_lib.class_dependencies((Foo, 1))

  def test_class_dependencies_recursive(self):
    self.assertEqual(
        schema_lib.class_dependencies(Node),
        [Node]
    )

  def test_class_dependencies_from_value(self):
    class Foo(pg.Object):
      x: int

    class Bar(pg.Object):
      y: str

    class A(pg.Object):
      foo: tuple[Foo, int]

    class B(pg.Object):
      pass

    class X(pg.Object):
      k: dict[str, B]

    class C(A):
      bar: Bar
      foo2: Foo | X

    a = A(foo=(Foo(1), 0))
    self.assertEqual(schema_lib.class_dependencies(a), [Foo, A, Bar, B, X, C])

    self.assertEqual(schema_lib.class_dependencies(1), [])


class SchemaPythonReprTest(unittest.TestCase):

  def assert_annotation(
      self,
      value_spec: pg.typing.ValueSpec,
      expected_annotation: str,
      strict: bool = False,
      **kwargs,
  ) -> None:
    self.assertEqual(
        schema_lib.annotation(value_spec, strict=strict, **kwargs),
        expected_annotation,
    )

  def test_annotation(self):
    # Bool.
    self.assert_annotation(pg.typing.Bool(), 'bool')
    self.assert_annotation(pg.typing.Bool().noneable(), 'bool | None')

    # Str.
    self.assert_annotation(pg.typing.Str(), 'str')
    self.assert_annotation(pg.typing.Str().noneable(), 'str | None')
    self.assert_annotation(pg.typing.Str(regex='a.*'), "str(regex='a.*')")
    self.assert_annotation(pg.typing.Str(regex='a.*'), "str(regex='a.*')")
    self.assert_annotation(
        pg.typing.Str(regex='a.*'), "pg.typing.Str(regex='a.*')", strict=True
    )

    # Int.
    self.assert_annotation(pg.typing.Int(), 'int')
    self.assert_annotation(pg.typing.Int().noneable(), 'int | None')
    self.assert_annotation(pg.typing.Int(min_value=0), 'int(min=0)')
    self.assert_annotation(pg.typing.Int(max_value=1), 'int(max=1)')
    self.assert_annotation(
        pg.typing.Int(min_value=0, max_value=1), 'int(min=0, max=1)'
    )

    self.assert_annotation(pg.typing.Int(), 'int', strict=True)
    self.assert_annotation(
        pg.typing.Int(min_value=0), 'pg.typing.Int(min_value=0)', strict=True
    )
    self.assert_annotation(
        pg.typing.Int(max_value=1), 'pg.typing.Int(max_value=1)', strict=True
    )
    self.assert_annotation(
        pg.typing.Int(min_value=0, max_value=1),
        'pg.typing.Int(min_value=0, max_value=1)',
        strict=True,
    )

    # Float.
    self.assert_annotation(pg.typing.Float(), 'float')
    self.assert_annotation(pg.typing.Float().noneable(), 'float | None')
    self.assert_annotation(pg.typing.Float(min_value=0), 'float(min=0)')
    self.assert_annotation(pg.typing.Float(max_value=1), 'float(max=1)')
    self.assert_annotation(
        pg.typing.Float(min_value=0, max_value=1), 'float(min=0, max=1)'
    )

    self.assert_annotation(pg.typing.Float(), 'float', strict=True)
    self.assert_annotation(
        pg.typing.Float(min_value=0),
        'pg.typing.Float(min_value=0)',
        strict=True,
    )
    self.assert_annotation(
        pg.typing.Float(max_value=1),
        'pg.typing.Float(max_value=1)',
        strict=True,
    )
    self.assert_annotation(
        pg.typing.Float(min_value=0, max_value=1),
        'pg.typing.Float(min_value=0, max_value=1)',
        strict=True,
    )

    # Enum
    self.assert_annotation(
        pg.typing.Enum[1, 'foo'].noneable(), "Literal[1, 'foo', None]"
    )

    # Object.
    self.assert_annotation(pg.typing.Object(Activity), 'Activity')
    self.assert_annotation(
        pg.typing.Object(Activity).noneable(), 'Activity | None'
    )
    self.assert_annotation(
        pg.typing.Object(Activity).noneable(), 'Activity | None',
        allowed_dependencies=set([Activity]),
    )
    self.assert_annotation(
        pg.typing.Object(Activity).noneable(), 'Any | None',
        allowed_dependencies=set(),
    )

    # List.
    self.assert_annotation(
        pg.typing.List(pg.typing.Object(Activity)), 'list[Activity]'
    )
    self.assert_annotation(
        pg.typing.List(pg.typing.Object(Activity)), 'list[Activity]',
        allowed_dependencies=set([Activity]),
    )
    self.assert_annotation(
        pg.typing.List(pg.typing.Object(Activity)), 'list[Any]',
        allowed_dependencies=set(),
    )
    self.assert_annotation(
        pg.typing.List(pg.typing.Object(Activity)).noneable(),
        'list[Activity] | None',
    )
    self.assert_annotation(
        pg.typing.List(pg.typing.Object(Activity).noneable()),
        'list[Activity | None]',
    )

    # Tuple.
    self.assert_annotation(
        pg.typing.Tuple([Activity, pg.typing.Str()]), 'tuple[Activity, str]'
    )
    self.assert_annotation(
        pg.typing.Tuple([Activity, pg.typing.Str()]), 'tuple[Activity, str]',
        allowed_dependencies=set([Activity]),
    )
    self.assert_annotation(
        pg.typing.Tuple([Activity, pg.typing.Str()]), 'tuple[Any, str]',
        allowed_dependencies=set(),
    )
    self.assert_annotation(
        pg.typing.Tuple([Activity, pg.typing.Str()]).noneable(),
        'tuple[Activity, str] | None',
    )

    # Dict.
    self.assert_annotation(
        pg.typing.Dict({'x': Activity, 'y': str}),
        '{\'x\': Activity, \'y\': str}'
    )
    self.assert_annotation(
        pg.typing.Dict({'x': Activity, 'y': str}),
        '{\'x\': Activity, \'y\': str}',
        allowed_dependencies=set([Activity]),
    )
    self.assert_annotation(
        pg.typing.Dict({'x': Activity, 'y': str}),
        '{\'x\': Any, \'y\': str}',
        allowed_dependencies=set(),
    )
    self.assert_annotation(
        pg.typing.Dict({'x': int, 'y': str}),
        'pg.typing.Dict({\'x\': int, \'y\': str})',
        strict=True,
    )
    self.assert_annotation(
        pg.typing.Dict(),
        'dict[str, Any]',
        strict=False,
    )

    class DictValue(pg.Object):
      pass

    self.assert_annotation(
        pg.typing.Dict([(pg.typing.StrKey(), DictValue)]),
        'dict[str, DictValue]',
        strict=False,
    )
    self.assert_annotation(
        pg.typing.Dict(),
        'dict[str, Any]',
        strict=True,
    )

    # Union.
    self.assert_annotation(
        pg.typing.Union(
            [pg.typing.Object(Activity), pg.typing.Object(Itinerary)]
        ).noneable(),
        'Union[Activity, Itinerary, None]',
    )
    self.assert_annotation(
        pg.typing.Union(
            [pg.typing.Object(Activity), pg.typing.Object(Itinerary)]
        ).noneable(),
        'Union[Activity, Any, None]',
        allowed_dependencies=set([Activity]),
    )

    # Any.
    self.assert_annotation(pg.typing.Any(), 'Any')
    self.assert_annotation(pg.typing.Any().noneable(), 'Any')

  def test_class_definition(self):
    self.assertEqual(
        schema_lib.class_definition(Activity, allowed_dependencies=set()),
        'class Activity:\n  description: str\n',
    )
    self.assertEqual(
        schema_lib.class_definition(Itinerary),
        inspect.cleandoc("""
            class Itinerary(Object):
              \"\"\"A travel itinerary for a day.\"\"\"
              day: int(min=1)
              type: Literal['daytime', 'nighttime']
              activities: list[Activity]
              # Hotel to stay if applicable.
              hotel: str(regex='.*Hotel') | None
            """) + '\n',
    )
    self.assertEqual(
        schema_lib.class_definition(
            PlaceOfInterest, allowed_dependencies=set()
        ),
        inspect.cleandoc("""
            class PlaceOfInterest:
              \"\"\"The name of a place of interest.

              A place of interest is a place that people ususally visit during their
              travels.
              \"\"\"
              name: str
            """) + '\n',
    )

    class A(pg.Object):
      pass

    self.assertEqual(
        schema_lib.class_definition(A, allowed_dependencies=set()),
        'class A:\n  pass\n',
    )
    self.assertEqual(
        schema_lib.class_definition(A),
        'class A(Object):\n  pass\n',
    )

    class C(pg.Object):
      x: str
      __kwargs__: typing.Any

    self.assertEqual(
        schema_lib.class_definition(C), 'class C(Object):\n  x: str\n'
    )

    class D(pg.Object):
      x: str
      @schema_lib.include_method_in_prompt
      def __call__(self, y: int) -> int:
        return len(self.x) + y

    self.assertEqual(
        schema_lib.class_definition(D),
        inspect.cleandoc(
            """
            class D(Object):
              x: str

              def __call__(self, y: int) -> int:
                return len(self.x) + y
            """) + '\n'
    )

    class E(pg.Object):
      x: str
      y: typing.Annotated[int, 'y', dict(exclude_from_prompt=True)]

    self.assertEqual(
        schema_lib.class_definition(E),
        inspect.cleandoc(
            """
            class E(Object):
              x: str
            """) + '\n'
    )

  def test_repr(self):
    class Foo(pg.Object):
      x: int

    @dataclasses.dataclass(frozen=True)
    class Bar:
      """Class Bar."""
      y: str

    @dataclasses.dataclass(frozen=True)
    class Baz(Bar):  # pylint: disable=unused-variable
      pass

    class A(pg.Object):
      foo: Foo

      @schema_lib.include_method_in_prompt
      def foo_value(self) -> int:
        return self.foo.x

      def baz_value(self) -> str:
        return 'baz'

    class B(A):
      bar: Bar
      foo2: Foo

      @schema_lib.include_method_in_prompt
      def bar_value(self) -> str:
        return self.bar.y

    schema = schema_lib.Schema([B])
    self.assertEqual(
        schema_lib.SchemaPythonRepr().class_definitions(schema),
        inspect.cleandoc('''
            class Foo:
              x: int

            class Bar:
              """Class Bar."""
              y: str

            class Baz(Bar):
              """Baz(y: str)"""
              y: str

            class B:
              foo: Foo
              bar: Bar
              foo2: Foo

              def bar_value(self) -> str:
                return self.bar.y

              def foo_value(self) -> int:
                return self.foo.x
            ''') + '\n',
    )

    self.assertEqual(
        schema_lib.SchemaPythonRepr().result_definition(schema), 'list[B]'
    )

    self.assertEqual(
        schema_lib.SchemaPythonRepr().repr(schema),
        inspect.cleandoc('''
            list[B]

            ```python
            class Foo:
              x: int

            class Bar:
              """Class Bar."""
              y: str

            class Baz(Bar):
              """Baz(y: str)"""
              y: str

            class B:
              foo: Foo
              bar: Bar
              foo2: Foo

              def bar_value(self) -> str:
                return self.bar.y

              def foo_value(self) -> int:
                return self.foo.x
            ```
            '''),
    )
    self.assertEqual(
        schema_lib.SchemaPythonRepr().repr(
            schema,
            include_result_definition=False,
            markdown=False,
        ),
        inspect.cleandoc('''
            class Foo:
              x: int

            class Bar:
              """Class Bar."""
              y: str

            class Baz(Bar):
              """Baz(y: str)"""
              y: str

            class B:
              foo: Foo
              bar: Bar
              foo2: Foo

              def bar_value(self) -> str:
                return self.bar.y

              def foo_value(self) -> int:
                return self.foo.x
            '''),
    )


class SchemaJsonReprTest(unittest.TestCase):

  def test_repr(self):
    schema = schema_lib.Schema([{'x': Itinerary}])
    self.assertEqual(
        schema_lib.SchemaJsonRepr().repr(schema),
        (
            '{"result": [{"x": {"_type": "Itinerary", "day":'
            ' int(min=1), "type": "daytime" | "nighttime", "activities":'
            ' [{"_type": "%s", "description": str}], "hotel":'
            ' str(regex=.*Hotel) | None}}]}' % (
                Activity.__type_name__,
            )
        ),
    )


class ValuePythonReprTest(unittest.TestCase):

  def test_repr(self):
    class Foo(pg.Object):
      x: int

    class A(pg.Object):
      foo: list[Foo]
      y: str | None

    self.assertEqual(
        schema_lib.ValuePythonRepr().repr(1, schema_lib.Schema(int)),
        '```python\n1\n```'
    )
    self.assertEqual(
        schema_lib.ValuePythonRepr().repr(lf.Template('hi, {{a}}', a='foo')),
        'hi, foo'
    )
    self.assertEqual(
        schema_lib.ValuePythonRepr().repr(
            A([Foo(1), Foo(2)], 'bar'), schema_lib.Schema(A), markdown=False,
        ),
        "A(foo=[Foo(x=1), Foo(x=2)], y='bar')",
    )
    self.assertEqual(
        schema_lib.ValuePythonRepr().repr(
            A([Foo(1), Foo(2)], 'bar'),
            schema_lib.Schema(A),
            markdown=True,
            compact=False,
            assign_to_var='output',
        ),
        inspect.cleandoc("""
            ```python
            output = A(
              foo=[
                Foo(
                  x=1
                ),
                Foo(
                  x=2
                )
              ],
              y='bar'
            )
            ```
            """),
    )
    self.assertEqual(
        schema_lib.ValuePythonRepr().repr(A),
        inspect.cleandoc("""
            ```python
            class Foo(Object):
              x: int

            class A(Object):
              foo: list[Foo]
              y: str | None
            ```
            """),
    )
    self.assertEqual(schema_lib.source_form(int), 'int')

  def test_parse(self):
    class Foo(pg.Object):
      x: int

    class A(pg.Object):
      foo: list[Foo]
      y: str | None

    self.assertEqual(
        schema_lib.ValuePythonRepr().parse(
            "A(foo=[Foo(x=1), Foo(x=2)], y='bar')", schema_lib.Schema(A)
        ),
        A([Foo(1), Foo(2)], y='bar'),
    )

  def test_parse_with_correction(self):
    class Foo(pg.Object):
      x: int

    class A(pg.Object):
      foo: list[Foo]
      y: str | None

    self.assertEqual(
        schema_lib.ValuePythonRepr().parse(
            "A(foo=[Foo(x=1), Foo(x=2)], y='bar'",
            schema_lib.Schema(A),
            autofix=1,
            autofix_lm=fake.StaticResponse(inspect.cleandoc("""
                    CorrectedCode(
                        corrected_code='A(foo=[Foo(x=1), Foo(x=2)], y=\\\'bar\\\')',
                    )
                    """)),
        ),
        A([Foo(1), Foo(2)], y='bar'),
    )

  def test_parse_class_def(self):
    self.assertTrue(
        inspect.isclass(
            schema_lib.ValuePythonRepr().parse(
                """
                class A:
                  x: Dict[str, Any]
                  y: Optional[Sequence[str]]
                  z: Union[int, List[int], Tuple[int]]
                """,
                permission=pg.coding.CodePermission.ALL,
            )
        )
    )


class ValueJsonReprTest(unittest.TestCase):

  def test_repr(self):
    self.assertEqual(schema_lib.ValueJsonRepr().repr(1), '{"result": 1}')

  def assert_parse(self, inputs, output) -> None:
    self.assertEqual(schema_lib.ValueJsonRepr().parse(inputs), output)

  def test_parse_basics(self):
    self.assert_parse('{"result": 1}', 1)
    self.assert_parse('{"result": "\\"}ab{"}', '"}ab{')
    self.assert_parse(
        '{"result": {"x": true, "y": null}}',
        {'x': True, 'y': None},
    )
    self.assert_parse(
        (
            '{"result": {"_type": "%s", "description": "play"}}'
            % Activity.__type_name__
        ),
        Activity('play'),
    )
    with self.assertRaisesRegex(
        schema_lib.JsonError, 'JSONDecodeError'
    ):
      schema_lib.ValueJsonRepr().parse('{"abc", 1}')

    with self.assertRaisesRegex(
        schema_lib.JsonError,
        'The root node of the JSON must be a dict with key `result`'
    ):
      schema_lib.ValueJsonRepr().parse('{"abc": 1}')

  def test_parse_with_surrounding_texts(self):
    self.assert_parse('The answer is {"result": 1}.', 1)

  def test_parse_with_new_lines(self):
    self.assert_parse(
        """
        {
            "result": [
"foo
bar"]
        }
        """,
        ['foo\nbar'])

  def test_parse_with_malformated_json(self):
    with self.assertRaisesRegex(
        schema_lib.JsonError, 'No JSON dict in the output'
    ):
      schema_lib.ValueJsonRepr().parse('The answer is 1.')

    with self.assertRaisesRegex(
        schema_lib.JsonError,
        'Malformated JSON: missing .* closing curly braces'
    ):
      schema_lib.ValueJsonRepr().parse('{"result": 1')


class ProtocolTest(unittest.TestCase):

  def test_schema_repr(self):
    self.assertIsInstance(
        schema_lib.schema_repr('json'), schema_lib.SchemaJsonRepr)
    self.assertIsInstance(
        schema_lib.schema_repr('python'), schema_lib.SchemaPythonRepr)
    with self.assertRaisesRegex(ValueError, 'Unsupported protocol'):
      schema_lib.schema_repr('text')

  def test_value_repr(self):
    self.assertIsInstance(
        schema_lib.value_repr('json'), schema_lib.ValueJsonRepr)
    self.assertIsInstance(
        schema_lib.value_repr('python'), schema_lib.ValuePythonRepr)
    with self.assertRaisesRegex(ValueError, 'Unsupported protocol'):
      schema_lib.value_repr('text')


class MissingTest(unittest.TestCase):

  def test_basics(self):
    a = Itinerary(
        day=1,
        type=schema_lib.Missing(),
        activities=schema_lib.Missing(),
        hotel=schema_lib.Missing(),
    )
    self.assertFalse(a.is_partial)
    self.assertEqual(str(schema_lib.Missing()), 'MISSING')
    self.assertEqual(str(a.type), "MISSING(Literal['daytime', 'nighttime'])")
    self.assertEqual(str(a.activities), 'MISSING(list[Activity])')
    self.assertEqual(str(a.hotel), "MISSING(str(regex='.*Hotel') | None)")

  def assert_missing(self, value, expected_missing):
    value = schema_lib.mark_missing(value)
    self.assertEqual(schema_lib.Missing.find_missing(value), expected_missing)

  def test_find_missing(self):
    self.assert_missing(
        Itinerary.partial(),
        {
            'day': schema_lib.MISSING,
            'type': schema_lib.MISSING,
            'activities': schema_lib.MISSING,
        },
    )

    self.assert_missing(
        Itinerary.partial(
            day=1, type='daytime', activities=[Activity.partial()]
        ),
        {
            'activities[0].description': schema_lib.MISSING,
        },
    )

  def test_mark_missing(self):
    class A(pg.Object):
      x: typing.Any

    self.assertEqual(schema_lib.mark_missing(1), 1)
    self.assertEqual(
        schema_lib.mark_missing(pg.MISSING_VALUE), pg.MISSING_VALUE
    )
    self.assertEqual(
        schema_lib.mark_missing(A.partial(A.partial(A.partial()))),
        A(A(A(schema_lib.MISSING))),
    )
    self.assertEqual(
        schema_lib.mark_missing(dict(a=A.partial())),
        dict(a=A(schema_lib.MISSING)),
    )
    self.assertEqual(
        schema_lib.mark_missing([1, dict(a=A.partial())]),
        [1, dict(a=A(schema_lib.MISSING))],
    )


class UnknownTest(unittest.TestCase):

  def test_basics(self):
    class A(pg.Object):
      x: int

    a = A(x=schema_lib.Unknown())
    self.assertFalse(a.is_partial)
    self.assertEqual(a.x, schema_lib.UNKNOWN)
    self.assertEqual(schema_lib.UNKNOWN, schema_lib.Unknown())


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Customized bounding box related to Gemini Object Detection."""

import json
import re
from typing import Any, Dict, List, Tuple, Union

from langfun.assistant.capabilities.gui import location
import pyglove as pg


class GeminiBBox(pg.Object):
  """Customized bounding box.

  Note: GeminiPro returns a unique JSON response structure for object detection
  tasks, which differs from the standard location.BBox format. To accommodate 
  this, we've created this custom class to adapt to GeminiPro's specific 
  response structure.
  """

  ymin: int
  xmin: int
  ymax: int
  xmax: int

  def scale(self,
            target_size: tuple[int, int],
            source_size: tuple[int, int] = (1000, 1000)) -> 'GeminiBBox':

    def _scale_x(v):
      return (v * target_size[0]) // source_size[0]

    def _scale_y(v):
      return (v * target_size[1]) // source_size[1]

    return GeminiBBox(
        ymin=_scale_y(self.ymin),
        xmin=_scale_x(self.xmin),
        ymax=_scale_y(self.ymax),
        xmax=_scale_x(self.xmax),
    )

  def resize(
      self, image_size: Tuple[int, int], resize_factor: int = 3
  ) -> 'GeminiBBox':
    """Resize a bounding box from its center.

    Args:
        image_size: A tuple (width, height) representing the size of the image.
        resize_factor: The factor by which to resize the bounding box. Defaults
          to 3.

    Returns:
        A bounding box with the expanded dimensions.
    """
    center_x = (self.xmin + self.xmax) // 2
    center_y = (self.ymin + self.ymax) // 2
    width = self.xmax - self.xmin
    height = self.ymax - self.ymin

    new_width = width * resize_factor
    new_height = height * resize_factor

    new_xmin = max(0, int(center_x - new_width // 2))
    new_ymin = max(0, int(center_y - new_height // 2))
    new_xmax = min(image_size[0], int(center_x + new_width // 2))
    new_ymax = min(image_size[1], int(center_y + new_height // 2))

    return GeminiBBox(
        xmin=new_xmin, ymin=new_ymin, xmax=new_xmax, ymax=new_ymax
    )

  def to_gui_bbox(self) -> location.BBox | None:
    try:
      return location.BBox(
          x=self.xmin, y=self.ymin, right=self.xmax, bottom=self.ymax
      )
    except AssertionError:
      # If the bounding box is not valid, return None.
      return None


def extract_json_candidate_from_text(raw_text: str) -> str:
  """Extracts a JSON candidate string from raw text."""
  # Try to find content within ```json ... ```
  match = re.search(r'```json\s*([\s\S]+?)\s*```', raw_text, re.IGNORECASE)
  if match:
    return match.group(1).strip()

  # Try to find content within ``` ... ```
  match = re.search(r'```\s*([\s\S]+?)\s*```', raw_text)
  if match:
    return match.group(1).strip()

  # If no code blocks, return the stripped raw text.
  return raw_text.strip()


def parse_and_convert_json(
    text: str, screen_size: Tuple[int, int] = (1000, 1000)
) -> Dict[str, location.BBox | None]:
  """Parse and convert json to bounding box.

  Args:
    text: The text to parse.
    screen_size: The screen size to scale the bounding box.

  Returns:
    A dictionary of bounding boxes.

  Example:
  >>> json_text = '```{"search button": [10, 20, 100, 200]}```'
  >>> bboxes = parse_and_convert_json(json_text, screen_size=(800, 600))
  >>> print(bboxes)
  {'search button': BBox(x=16, y=6, right=160, bottom=60)}
  """

  def parse_json(t: str) -> Union[Dict[str, str], List[str], None]:
    """Parse text to json."""
    try:
      return json.loads(t)
    except json.JSONDecodeError:
      return None

  def can_cast_to_int(obj: Any) -> bool:
    try:
      int(obj)
      return True
    except (ValueError, TypeError):
      return False

  def convert_to_bbox(
      data: Union[Dict[str, Any], List[Any], None], screen_size: Tuple[int, int]
  ) -> Dict[str, location.BBox | None]:
    """Convert data to bounding box."""
    result = {}
    if not data:
      return result

    if isinstance(data, list):
      if len(data) == 4 and all(can_cast_to_int(item) for item in data):
        return {
            'element': (
                GeminiBBox(
                    xmin=int(data[1]),
                    ymin=int(data[0]),
                    xmax=int(data[3]),
                    ymax=int(data[2]),
                )
                .scale(screen_size)
                .to_gui_bbox()
            )
        }
      for item in data:
        if isinstance(item, dict):
          result.update(convert_to_bbox(item, screen_size))
      return result

    for key, value in data.items():
      if not value:
        continue

      if isinstance(value, list):
        if len(value) == 4 and all(can_cast_to_int(item) for item in value):
          result[key] = (
              GeminiBBox(
                  xmin=int(value[1]),
                  ymin=int(value[0]),
                  xmax=int(value[3]),
                  ymax=int(value[2]),
              )
              .scale(screen_size)
              .to_gui_bbox()
          )
      elif isinstance(value, dict):
        result.update(convert_to_bbox(value, screen_size))
    return result

  parsed_data = parse_json(extract_json_candidate_from_text(text))
  if parsed_data is None:
    return {}

  return convert_to_bbox(parsed_data, screen_size)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Evaluation experiment runners."""
import abc
import collections
import concurrent.futures
import random
import threading
import time
import traceback
from typing import Any, Annotated, Callable, Iterator

from langfun import core as lf
from langfun.core.eval.v2 import checkpointing
from langfun.core.eval.v2 import evaluation as evaluation_lib
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
from langfun.core.eval.v2 import progress_tracking
from langfun.core.eval.v2 import reporting
from langfun.core.llms.cache import in_memory
import pyglove as pg

Runner = experiment_lib.Runner
Example = example_lib.Example
Evaluation = evaluation_lib.Evaluation
Experiment = experiment_lib.Experiment
Plugin = experiment_lib.Plugin


_RUN_MANIFEST = 'run.json'


class RunnerBase(Runner):
  """A simple runner that runs evaluations and their examples sequentially."""

  tqdm: Annotated[
      bool,
      (
          'If True, force using tqdm for progress update. Otherwise, determine '
          'it automatically based on the running environment (console vs. '
          'notebook)'
      )
  ] = False

  plugins = [
      checkpointing.BulkCheckpointer(),
      reporting.HtmlReporter(),
  ]

  def _on_bound(self):
    super()._on_bound()

    # Install the tqdm plugin if needed.
    with pg.notify_on_change(False):
      self.plugins.append(progress_tracking.progress_tracker(self.tqdm))

    self._io_pool_lock = threading.Lock()
    self._io_pool = concurrent.futures.ThreadPoolExecutor(max_workers=16)
    # TODO(daiyip): render background errors.
    self._background_last_error = None

  def background_run(self, func: Callable[..., Any], *args, **kwargs) -> None:
    """Runs the function with the IO pool."""
    def _background_run(*args, **kwargs):
      try:
        func(*args, **kwargs)
      except Exception as e:  # pylint: disable=broad-except
        self._background_last_error = e

    with self._io_pool_lock:
      if self._io_pool is not None:
        self._io_pool.submit(_background_run, *args, **kwargs)

  def _all_plugins(self, experiment: Experiment) -> Iterator[Plugin]:
    """Returns all plugins for the experiment."""
    for plugin in self.plugins:
      yield plugin
    for plugin in experiment.plugins:
      yield plugin

  #
  # IO operations for saving running files.
  #

  def _save_run_manifest(self) -> None:
    def _save():
      pg.symbolic.deref(self.current_run.clone(), recursive=True).save(
          self.current_run.output_path_for(
              self.current_run.experiment, _RUN_MANIFEST
          ),
          hide_default_values=True
      )
    self.background_run(_save)

  def on_run_start(self) -> None:
    """Called when a runner is started."""
    self._save_run_manifest()

    for plugin in self._all_plugins(self.current_run.experiment):
      plugin.on_run_start(self, self.current_run.experiment)

  def on_run_complete(self) -> None:
    """Called when a runner is complete."""
    for plugin in self._all_plugins(self.current_run.experiment):
      plugin.on_run_complete(self, self.current_run.experiment)

  def on_run_abort(self, error: Exception) -> None:
    """Called when a runner is aborted."""
    for plugin in self._all_plugins(self.current_run.experiment):
      plugin.on_run_abort(self, self.current_run.experiment, error)

  def on_experiment_start(self, experiment: Experiment) -> None:
    """Called when an evaluation is started."""
    # Start the progress of the evaluation.
    num_examples_to_evaluate = 0
    if experiment.is_leaf:
      assert isinstance(experiment, Evaluation)
      num_examples_to_evaluate = (
          len(self.current_run.example_ids)
          if self.current_run.example_ids else experiment.num_examples
      )
      experiment.progress.start(total=num_examples_to_evaluate)
    else:
      experiment.progress.start(total=len(experiment.leaf_nodes))

    # Notify the plugins of the experiment start.
    for plugin in self._all_plugins(experiment):
      plugin.on_experiment_start(self, experiment)

    if experiment.is_leaf:
      experiment.info(
          f'Starting evaluation {experiment.id!r} with '
          f'{num_examples_to_evaluate} examples to evaluate.'
      )

  def on_experiment_skipped(self, experiment: Experiment) -> None:
    """Called when an evaluation is skipped."""
    # Skip event will only be triggered for leaf evaluations.
    assert experiment.is_leaf
    experiment.progress.start(total=1)
    experiment.progress.increment_skipped(1)

    if experiment.is_leaf:
      experiment.info(
          f'Evaluation {experiment.id!r} is skipped.'
      )

    # Notify the plugins of the experiment skip.
    for plugin in self._all_plugins(experiment):
      plugin.on_experiment_skipped(self, experiment)

    # Only leaf evaluations will trigger the complete notification of the
    # ancestors.
    self._update_ancestor_progresses(experiment)

  def on_experiment_complete(self, experiment: Experiment) -> None:
    """Called when an evaluation is complete."""
    progress = experiment.progress
    progress.stop()

    # Notify the plugins of the experiment complete.
    for plugin in self._all_plugins(experiment):
      plugin.on_experiment_complete(self, experiment)

    # Only leaf evaluations will trigger the complete notification of the
    # ancestors.
    if experiment.is_leaf:
      self._update_ancestor_progresses(experiment)
      self._log_experiment_completion(experiment)

  def _log_experiment_completion(self, experiment: Experiment):
    example_ids = (
        self.current_run.example_ids if self.current_run.example_ids else
        list(range(1, experiment.num_examples + 1))
    )
    num_from_checkpoint, num_processed = 0, 0
    for example_id in example_ids:
      status = experiment.state.get_status(example_id)
      if status.newly_processed:
        num_processed += 1
      else:
        num_from_checkpoint += 1
    experiment.info(
        f'{experiment.id} completed with {num_from_checkpoint + num_processed} '
        f'examples evaluated ({num_from_checkpoint} from checkpoint, '
        f'{num_processed} newly processed).'
    )

  def on_experiment_abort(
      self, experiment: Experiment, error: BaseException) -> None:
    """Called when an evaluation is complete."""
    assert experiment.is_leaf
    experiment.fatal(f'{error}\n\n{traceback.format_exc()}')

    # Notify the plugins of the experiment abort.
    for plugin in self._all_plugins(experiment):
      plugin.on_experiment_abort(self, experiment, error)

  def _update_ancestor_progresses(self, experiment: Experiment):
    """Updates the progresses of the parent nodes of the experiment."""
    parent = experiment.parent
    progress = experiment.progress
    while parent is not None:
      parent_progress = parent.progress
      if progress.is_failed:
        parent_progress.increment_failed()
      elif progress.is_skipped:
        parent_progress.increment_skipped()
      else:
        # A evaluation could be considered as done if it has processed all the
        # examples specified by `example_ids`.
        assert progress.is_completed
        parent_progress.increment_processed()

      if parent_progress.is_completed:
        self.on_experiment_complete(parent)
      elif parent_progress.is_skipped:
        self.on_experiment_skipped(parent)
      parent = parent.parent

  def on_example_start(
      self,
      experiment: Experiment,
      example: Example
  ) -> None:
    """Called when an evaluation example is started."""
    for plugin in self._all_plugins(experiment):
      plugin.on_example_start(self, experiment, example)
    experiment.info(f'Starting to evaluate example {example.id}.')

  def on_example_complete(
      self,
      experiment: Experiment,
      example: Example
  ) -> None:
    """Called when an evaluation example is complete."""
    if example.newly_processed:
      if example.error is None:
        experiment.progress.increment_processed()
        experiment.info(
            f'Example {example.id} is successfully evaluated (with process) in '
            f'{example.elapse:.2f} seconds.'
        )
      else:
        experiment.progress.increment_failed()
        experiment.error(
            (
                f'Failed to evaluate example {example.id} in'
                f'{example.elapse:.2f} seconds.'
            ),
            error=example.error
        )
    else:
      experiment.progress.increment_skipped()
      experiment.info(
          f'Example {example.id} is successfully evaluated (without reprocess) '
          f'in {example.elapse:.2f} seconds.'
      )

    experiment.usage_summary.merge(example.usage_summary)
    experiment.progress.update_execution_summary(example.execution_status)

    parent = experiment.parent
    while parent is not None:
      parent.usage_summary.merge(example.usage_summary)
      parent = parent.parent

    for plugin in self._all_plugins(experiment):
      plugin.on_example_complete(self, experiment, example)

  def run(self) -> None:
    """Runs the experiment."""
    # Resets the experiment before getting start.
    for node in self.current_run.experiment.nodes:
      node.reset()

    # Start the run.
    self.on_run_start()
    cache = None

    try:
      # Start the non-leaf nodes.
      for node in self.current_run.experiment.nonleaf_nodes:
        self.on_experiment_start(node)

      # Skip evaluations if needed.
      if self.current_run.filter is not None:
        targets = []
        for evaluation in self.current_run.experiment.leaf_nodes:
          if self.current_run.filter(evaluation):
            targets.append(evaluation)
          else:
            self.on_experiment_skipped(evaluation)
      else:
        targets = self.current_run.experiment.leaf_nodes

      # Prepare the global cache if needed.
      global_settings = {}
      if self.current_run.use_cache == 'global':
        cache = self._load_or_create_cache(self.current_run.experiment)
        global_settings['cache'] = cache

      # Evaluate the leaf evaluations if not skipped.
      with lf.use_settings(**global_settings):
        self._run(targets)

      self.on_run_complete()
    except Exception as e:  # pylint: disable=broad-except
      self.on_run_abort(e)
      raise e
    finally:
      if cache is not None:
        self.background_run(cache.save)

      # Wait for the background tasks to finish.
      with self._io_pool_lock:
        self._io_pool, io_pool = None, self._io_pool
      io_pool.shutdown(wait=True)

  @abc.abstractmethod
  def _run(self, evaluations: list[Evaluation]) -> None:
    """Runs multiple evaluations."""

  def run_evaluation(self, evaluation: Evaluation) -> None:
    """Runs the evaluation."""
    try:
      self.on_experiment_start(evaluation)

      per_evaluation_settings = {}
      cache = None
      if self.current_run.use_cache == 'per_dataset':
        cache = self._load_or_create_cache(evaluation)
        per_evaluation_settings['cache'] = cache

      with lf.use_settings(**per_evaluation_settings):
        if self.current_run.example_ids is None:
          items = (
              Example(id=i + 1, input=ex) for i, ex in enumerate(
                  evaluation.example_inputs)
          )
        else:
          items = (
              Example(
                  id=example_id,
                  input=evaluation.example_input_by_id(example_id)
              ) for example_id in self.current_run.example_ids
          )
        if self.current_run.shuffle_inputs:
          items = list(items)
          random.shuffle(items)
        self._evaluate_items(evaluation, items)

      if cache:
        self.background_run(cache.save)
      self.on_experiment_complete(evaluation)
    except BaseException as e:  # pylint: disable=broad-except
      self.on_experiment_abort(evaluation, e)
      raise e

  @abc.abstractmethod
  def _evaluate_items(
      self, evaluation: Evaluation, items: Iterator[Example]
  ) -> None:
    """Evaluates the items of an evaluation."""

  def evaluate_item(
      self,
      evaluation: Evaluation,
      item: Example
  ) -> Example:
    """Runs the evaluation example."""
    self.on_example_start(evaluation, item)
    item = evaluation.evaluate(
        item,
        raise_if_has_error=self.current_run.raise_if_has_error,
        reevaluate_upon_previous_errors=self.current_run.reevaluate_upon_previous_errors,
    )
    self.on_example_complete(evaluation, item)
    return item

  def _load_or_create_cache(self, experiment: Experiment) -> lf.LMCache | None:
    """Loads or creates the cache."""
    return in_memory.InMemory(
        self.current_run.output_path_for(experiment, 'cache.json')
    )


class SequentialRunner(RunnerBase):
  """Sequential runner.

  Sequential runner runs all evaluations and their examples in sequence,
  as well as the background tasks, it allows the developer to catch all
  exceptions thrown from the background tasks, making it easier to debug.
  """

  NAME = 'sequential'

  def background_run(
      self, func: Callable[..., Any], *args: Any, **kwargs: Any
  ) -> None:
    """Runs the function with the IO pool."""
    func(*args, **kwargs)

  def _run(self, evaluations: list[Evaluation]) -> None:
    """Runs the experiment in sequence."""
    for e in evaluations:
      self.run_evaluation(e)

  def _evaluate_items(
      self, evaluation: Evaluation, items: Iterator[Example]
  ) -> None:
    """Runs the evaluation items in sequence."""
    for item in items:
      self.evaluate_item(evaluation, item)


class DebugRunner(SequentialRunner):
  """Debug runner."""

  NAME = 'debug'

  # Do not use the checkpointer for debug runner.
  plugins = []

  def _on_bound(self):
    super()._on_bound()
    if self.current_run.example_ids is None:
      self.current_run.rebind(example_ids=[1], skip_notification=True)
    self.current_run.rebind(raise_if_has_error=True, skip_notification=True)

  def _save_run_manifest(self) -> None:
    """Do nothing to avoid overriden existing runs."""


class ParallelRunner(RunnerBase):
  """Parallel runner."""

  NAME = 'parallel'

  timeout: Annotated[
      int | None,
      'Timeout for each evaluation example.'
  ] = None

  concurrent_startup_delay: Annotated[
      tuple[int, int] | None,
      (
          'A range of seconds to delay the initial evaluation of each thread '
          'in the thread pool, helping to prevent a burst in LLM QPS at '
          'startup. If set to None, no delay will be applied.'
      )
  ] = None

  def _run(self, evaluations: list[Evaluation]) -> None:
    """Runs the evaluations in parallel."""
    def _run_group(evaluation_group: list[Evaluation]):
      for e in evaluation_group:
        self.run_evaluation(e)

    # Run evaluations in parallel groupped by resource key.
    groups: dict[str, list[Evaluation]] = collections.defaultdict(list)
    for e in evaluations:
      resource_ids = e.resource_ids()
      if not resource_ids:
        group_id = e.id
      else:
        # TODO(daiyip): support group that requires multiple resources.
        group_id = resource_ids.pop()
      groups[group_id].append(e)

    for _, _, _ in lf.concurrent_map(
        _run_group,
        groups.values(),
        max_workers=max(64, len(groups)),
        timeout=self.timeout,
        silence_on_errors=None,
    ):
      pass

  def _evaluate_items(
      self, evaluation: Evaluation, items: Iterator[Example]
  ) -> None:
    """Override run items to run in parallel."""
    if self.concurrent_startup_delay is not None:
      thread_delayed = {}
      def _evaluate_item(item: Example):
        thread_id = threading.current_thread().ident
        if thread_id not in thread_delayed:
          thread_delayed[thread_id] = True
          time.sleep(random.randint(*self.concurrent_startup_delay))
        return self.evaluate_item(evaluation, item)
    else:
      def _evaluate_item(item: Example):
        return self.evaluate_item(evaluation, item)

    for _, _, _ in lf.concurrent_map(
        _evaluate_item,
        items,
        max_workers=evaluation.max_workers,
        timeout=self.timeout,
        silence_on_errors=None,
    ):
      pass



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun Component."""

from typing import ContextManager
import pyglove as pg


# Default value marker that indicates to raise error.
RAISE_IF_HAS_ERROR = (pg.MISSING_VALUE,)


class Component(pg.ContextualObject):
  """Base class for langfun components."""

  # Allow symbolic assignment, which invalidates the object and recomputes
  # states upon update.
  allow_symbolic_assignment = True

  # Class property that indicates whether to use `sym_eq` for `__eq__`,
  # `sym_ne` for `__ne__`, and `sym_hash` for `__hash__`.
  use_symbolic_comparison = False

  def __init_subclass__(cls):
    super().__init_subclass__()

    # Find class attributes that do not have annotations but are `Component`,
    # and treat them as symbolic fields.
    additional_fields = []
    for attr_name in dir(cls):
      if (
          attr_name.startswith('_')
          or attr_name.isupper()
          or attr_name in cls.__schema__.fields
      ):
        continue
      attr_value = getattr(cls, attr_name)
      if isinstance(attr_value, pg.Inferentiable):
        value_spec = pg.typing.Any()
      elif isinstance(attr_value, Component):
        value_spec = pg.typing.Object(Component)
      else:
        value_spec = None
      if value_spec:
        field = pg.typing.create_field((attr_name, value_spec))
        field.value.set_default(attr_value)
        additional_fields.append(field)
    if additional_fields:
      cls.update_schema(additional_fields)


# Aliases from PyGlove for ease of access.
context = pg.contextual_override
get_contextual_override = pg.utils.get_contextual_override
context_value = pg.utils.contextual_value
all_contextual_values = pg.utils.all_contextual_values
contextual = pg.contextual_attribute

# Decorator for setting the positional arguments for Component.
use_init_args = pg.use_init_args


def use_settings(
    *,
    cascade: bool = False,
    **settings,
) -> ContextManager[dict[str, pg.utils.ContextualOverride]]:
  """Shortcut method for overriding component attributes.

  Args:
    cascade: If True, this override will apply to both current scope and nested
      scope, meaning that this `lf.context` will take precedence over all
      nested `lf.context` on the overriden variables.
    **settings: Key/values as override for component attributes.

  Returns:
    A dict of attribute names to their contextual overrides.
  """
  return context(cascade=cascade, override_attrs=True, **settings)
